#!/usr/bin/env python3

# copied on 5 May 2021, from https://git.ligo.org/pe/O3/o3a_catalog_events/-/blob/master/scripts/make_uni_comov_skymap.py


#import pesummary
#from pesummary.gw.file.read import read
import argparse
import logging
from logging import info, error, warn
from lalinference.nest2pos import draw_posterior
import h5py
import numpy as np
import astropy
from astropy.cosmology import LambdaCDM
Planck15_lal = LambdaCDM(H0=67.90, Om0=0.3065, Ode0=0.6935)

parser = argparse.ArgumentParser('Program to resample lalinference posteriors from euclidean to uniform-in-comoving-volume distance prior')
parser.add_argument('--runid',help='RunID to use from file. If not given, will apply to all runs',default=None)
parser.add_argument('-v','--verbose',default=False, action='store_true',help='Produce verbose output')
parser.add_argument('--sfr-power',help='Provide a value X such that the reweighting includes a star formation rate evolution as (1+z)^X (default=0, no evolution)', default=0, metavar='X')
parser.add_argument('input',help='Input file to process', metavar='input.h5')
parser.add_argument('output',help='PESummary file to output after resampling',metavar='output.h5', default=None)


def resample(indata, verbose=False, lcdm=Planck15_lal, sfr_power=0):
    """
    Resample the input data with the given weight
    indata: PESummary samples object
    """
    # Resampling factor is (dV_c/dz * dz/ddl * ddl/dV_E ) * (dt_src / dt_det)
    dl = np.array(indata['luminosity_distance'])
    Mpc = astropy.units.megaparsec
    sr = astropy.units.steradian
    D_H = (astropy.constants.c / lcdm.H0).to_value(unit=Mpc)
    try:
        z=np.array(indata['redshift'])
    except KeyError:
        z=np.array([astropy.cosmology.z_at_value(lcdm.luminosity_distance, this_dl*Mpc) for this_dl in dl])
    E = lcdm.efunc(z)
    weight = 1.0/((1+z)**(2.0-sfr_power)*(E*(dl/D_H) + (1+z)**2))
    return draw_posterior(indata, np.log(weight), verbose=verbose)


def main(inputfile, outputfile, runid=None, verbose=False, sfr_power=0):
    infile = h5py.File(inputfile, 'r')
    
    if runid and runid not in infile:
        raise KeyError(f"Input file {inputfile} does not contain table {runid}")

    if runid is not None:
        tabnames = [runid]
    else:
        tabnames = set(infile.keys())
    info(f"Will process {tabnames} from {inputfile}")


    # Create new output file object
    outfile = h5py.File(outputfile, 'w')

    # For each run to process, perform reweighting
    for t in tabnames:
        #Copy structure for configs etc
        infile.copy(infile[t],outfile)
        if not 'posterior_samples' in list(infile[t]):
            continue
        oldpos = infile[t]['posterior_samples']
        newpos = resample(oldpos, verbose=verbose, sfr_power=sfr_power)
        info(f"Drew {len(newpos)}/{len(oldpos)} samples for {t}")
    
        # Add resampled run to output file
        del outfile[t]['posterior_samples']
        outfile[t]['posterior_samples']=newpos

        # Apply resampling to priors too, if they exist
        try:
            oldprior = infile[t]['priors']['samples']
            newprior = resample(oldprior, verbose=verbose, sfr_power=sfr_power)
            outfile[t]['priors']['samples']=newprior
        except (KeyError, ValueError, AttributeError):
            info(f"No prior samples found for {t}")
        except NotImplementedError:
            warn(f"Warning: cannot reweight prior samples as they have no redshift")

    # Save and exit
    infile.close()
    outfile.close()

if __name__=='__main__':
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.INFO)
    main(args.input, args.output, runid=args.runid, sfr_power=args.sfr_power)




#! /usr/bin/env python
"""
util_MassGrid.py <fnames>
    - prints out <m1> <m2> <lnL>  <neff> \sigma/L for each individual ILE *mass point*, to stdout
    - fnames can be an arbitrary mix of .sqlite or .xml files
    - currently *hardcodes* intrinsic (mass) parameter list

Used to generate ascii dumps summarizing runs
"""
import sys
import os
import RIFT.misc.xmlutils as xmlutils
from optparse import OptionParser
from igwn_ligolw import lsctables, table, utils

import numpy as np
import RIFT.misc.weight_simulations as weight_simulations

# optp= OptionParser()
# opts, args = optp.parse_args()

data_at_intrinsic = {}
for fname in sys.argv[1:]:
    fExtension = fname.split(".")[-1]
    # if .sqlite extension
    if fExtension == "sqlite":
        samples = xmlutils.db_to_samples(fname, lsctables.SnglInspiralTable, ['mass1', 'mass2', 'snr', 'tau0', 'event_duration'])
        for row in samples:  # print each individual row, don't reweight yet
            if not (data_at_intrinsic.has_key((row.mass1,row.mass2))):
                data_at_intrinsic[(row.mass1,row.mass2)] = [[row.snr, row.tau0, row.event_duration]]
            else:
                data_at_intrinsic[(row.mass1,row.mass2)].append([row.snr, row.tau0, row.event_duration])
    # otherwise xml: could have just used ligolw_print
    elif fExtension == "xml" or fExtension == "gz":
        samples = table.get_table(utils.load_filename(fname), lsctables.SnglInspiralTable.tableName)
        for row in samples:  # print each individual row, don't reweight yet
            if not (data_at_intrinsic.has_key((row.mass1,row.mass2))):
                data_at_intrinsic[(row.mass1,row.mass2)] = [[row.snr, row.tau0, row.event_duration]]
            else:
                data_at_intrinsic[(row.mass1,row.mass2)].append([row.snr, row.tau0, row.event_duration])


# Loop over each key and print out
for intrinsic in data_at_intrinsic.keys():
    lnL, neff, sigmaOverL =   np.transpose(data_at_intrinsic[intrinsic])
    lnLmax = np.max(lnL)
    sigma = sigmaOverL*np.exp(lnL-lnLmax)  # remove overall Lmax factor, which factors out from the weights constructed from \sigma
    wts = weight_simulations.AverageSimulationWeights(None, None,sigma)   
    lnLmeanMinusLmax = np.log(np.sum(np.exp(lnL - lnLmax)*wts))
    sigmaNetOverL = (np.sqrt(1./np.sum(1./sigma/sigma)))/np.exp(lnLmeanMinusLmax)
    print(intrinsic[0], intrinsic[1], lnLmeanMinusLmax+lnLmax, np.sum(neff), sigmaNetOverL)





# Test script for comparing GMM integrator to existing mcsampler integrator in
# RIFT. A simple n-dimensional integrand consisting of a highly-correlated
# Gaussian is used.

from __future__ import print_function
import numpy as np
from scipy.stats import multivariate_normal
from scipy.stats import truncnorm
import matplotlib.pyplot as plt

from RIFT.integrators import mcsampler, mcsamplerEnsemble, mcsamplerGPU, mcsamplerNFlow
from RIFT.integrators import mcsamplerPortfolio

import optparse
parser = optparse.OptionParser()
parser.add_option("--n-max",type=int,default=40000)
parser.add_option("--n-eff",type=int,default=1000)
parser.add_option("--use-lnL",action='store_true')
parser.add_option("--save-plot",action='store_true')
parser.add_option("--as-test",action='store_true')
parser.add_option("--no-adapt",action='store_true')
parser.add_option("--floor-level",default=0.0,type=float)
parser.add_option("--n-chunk",default=10000,type=int)
parser.add_option("--verbose",action='store_true')
opts, args = parser.parse_args()

save_fairdraws=False
save_fairdraw_prefix="fairdraw_demo"


verbose=opts.verbose

tempering_exp =0.1

### test parameters

# width of domain of integration, same for all dimensions
width = 10.0                                                    
# number of dimensions
ndim = 3                                                        
# mean of the Gaussian, allowed to occupy middle half of each dimension
mu = np.random.uniform(-1 * width / 4.0, width / 4.0, ndim)    
# max number of samples for mcsampler
nmax = opts.n_max                                         
# number of iterations for mcsamplerEnsemble
n_iters = int(nmax/opts.n_chunk)

llim = -1 * width / 2
rlim = width / 2

### generate list of named parameters
params = [str(i) for i in range(ndim)]

### generate the covariance matrix
cov = np.identity(ndim)
cov[ndim - 1][ndim - 1] = 0.05 # make it narrower in one dimension

### add some covariance (to test handling of strongly-correlated likelihoods)
cov[0][ndim - 1] = -0.1
cov[ndim - 1][0] = -0.1

### define integrand as a weighted sum of Gaussians
scale_factor = 100

def f(x1, x2, x3):
    x = np.array([x1, x2, x3]).T
    return scale_factor*multivariate_normal.pdf(x, mu, cov)
def ln_f(x1, x2, x3):
    x = np.array([x1, x2, x3]).T
    return np.log(scale_factor*multivariate_normal.pdf(x, mu, cov)+1e-100)

### initialize samplers
sampler = mcsampler.MCSampler()
samplerEnsemble = mcsamplerEnsemble.MCSampler()
samplerEnsemble_v2 = mcsamplerEnsemble.MCSampler()
samplerAC = mcsamplerGPU.MCSampler()
samplerAC_v2 = mcsamplerGPU.MCSampler()
samplerAC_v3 = mcsamplerGPU.MCSampler()
samplerAC_v4 = mcsamplerGPU.MCSampler()
samplerNF = mcsamplerNFlow.MCSampler()
portfolio_breakpoints = None
portfolio_list = [samplerAC_v2,samplerAC_v3]
#portfolio_list = [samplerAC_v2,samplerEnsemble_v2]
#portfolio_list = [samplerAC_v2,samplerEnsemble_v2, samplerAV_v2]
#portfolio_breakpoints = np.array([0,2,5])
#portfolio_list = [samplerAC_v2,samplerAC_v3,samplerAC_v4, samplerEnsemble_v2]
samplerPortfolio = mcsamplerPortfolio.MCSampler(portfolio=portfolio_list,portfolio_freeze_wt=0.1)  # use AC as example of portfolio

### add parameters
for p in params:
    sampler.add_parameter(p, np.vectorize(lambda x:1/(rlim-llim)), 
            prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
            left_limit=llim, right_limit=rlim,
            adaptive_sampling=not opts.no_adapt)
    samplerEnsemble.add_parameter(p, 
                                  pdf=np.vectorize(lambda x:1/(rlim-llim)),
                                  prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
                                  left_limit=llim, right_limit=rlim,adaptive_sampling=not opts.no_adapt)
    # for AC sampler, make sure pdf and prior pdfs are *normalized* *initially*
    samplerAC.add_parameter(p, pdf=np.vectorize(lambda x:1/(rlim-llim)),
            prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
            left_limit=llim, right_limit=rlim,adaptive_sampling=not opts.no_adapt)
    samplerNF.add_parameter(p, pdf=np.vectorize(lambda x:1/(rlim-llim)),
            prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
            left_limit=llim, right_limit=rlim,adaptive_sampling=True)

    samplerPortfolio.add_parameter(p, pdf=np.vectorize(lambda x:1/(rlim-llim)),
            prior_pdf=np.vectorize(lambda x:1/(rlim-llim)),
            left_limit=llim, right_limit=rlim,adaptive_sampling=True)



# number of Gaussian components to use in GMM
n_comp = 1

### integrate
extra_args = {"n": opts.n_chunk,"n_adapt":100, "floor_level":opts.floor_level,"tempering_exp" :tempering_exp,"neff":opts.n_eff}  # don't terminate
integral_1, var_1, eff_samp_1, _ = sampler.integrate(f, *params, 
        no_protect_names=True, nmax=nmax, save_intg=True,verbose=verbose,**extra_args)
print(" default {} {} {} ".format(integral_1, np.sqrt(var_1)/integral_1, eff_samp_1))
print(" --- finished default --")
use_lnL = opts.use_lnL
return_lnI=opts.use_lnL
if use_lnL:
    infunc = ln_f
else:
    infunc = f
integral_1b, var_1b, eff_samp_1b, _ = samplerAC.integrate(infunc, *params, 
        no_protect_names=True, nmax=nmax, save_intg=True,verbose=verbose,use_lnL=use_lnL,**extra_args)
if use_lnL:
    rel_error = np.exp(var_1b/2 - integral_1b) # I think ...
    integral_1b = np.exp(integral_1b)
else:
    rel_error = np.sqrt(var_1b)/integral_1b
print(" AC {} {} {} ".format(integral_1b, rel_error, eff_samp_1b))
print(" --- finished AC --")
integral_2, var_2, eff_samp_2, _ = samplerEnsemble.integrate(infunc, *params, 
        min_iter=n_iters, max_iter=n_iters, correlate_all_dims=True, n_comp=n_comp,super_verbose=False,verbose=verbose,use_lnL=use_lnL,return_lnI=return_lnI,**extra_args)
if return_lnI and use_lnL:
    rel_error_2 = np.exp(var_2/2 - integral_2)
    integral_2 = np.exp(integral_2)
else:
    rel_error_2 = np.sqrt(var_2)/integral_2
print(" GMM {} {} {} ".format(integral_2, rel_error_2, eff_samp_2))
print(" --- finished GMM --")
samplerNF.setup()
integral_3, var_3, eff_samp_3, _ = samplerNF.integrate_log(ln_f, *params, no_protect_names=True,
        nmax=nmax,
        min_iter=n_iters, max_iter=n_iters, **extra_args)
integral_3 = np.exp(integral_3)
var_3 = np.exp(var_3)
print(" -- finished NF -- ")

samplerPortfolio.setup(portfolio_breakpoints=portfolio_breakpoints) # vanilla setup for all members, nothing special. portfolio_breakpoints sets active list
# Set special optoins for one of them
samplerAC_v2.setup(n_bins=2) # force fairly low efficiency
samplerAC_v3.setup(n_bins=10) #
samplerEnsemble_v2.setup(n=1e4) # need larger step size : problem of not making enough intermediate points

integral_4, var_4, eff_samp_4, _ = samplerPortfolio.integrate_log(ln_f, *params, no_protect_names=True,
        nmax=nmax,
        min_iter=n_iters, max_iter=n_iters, verbose=True, **extra_args)
integral_4 = np.exp(integral_4)
var_4 = np.exp(var_4)
print(" -- finished portfolio -- ")

print(np.array([integral_1,integral_1b,integral_2,integral_3,integral_4])*width**3)  # remove prior factor, should get result of normal over domain
print(np.array([np.sqrt(var_1)/integral_1,np.sqrt(var_1b)/integral_1b,np.sqrt(var_2)/integral_2,np.sqrt(var_3)/integral_3]))  # relative error in each integral
print(" AC/default ",  integral_1b/integral_1, np.sqrt(var_1)/integral_1)  # off by width**3
print(" GMM/default ",integral_2/integral_1, np.sqrt(var_1)/integral_1, np.sqrt(var_2)/integral_2)
print(" AV/default ",integral_3/integral_1, np.sqrt(var_1)/integral_1, np.sqrt(var_3)/integral_3)
print(" Portfolio/default ",integral_4/integral_1, np.sqrt(var_1)/integral_1, np.sqrt(var_4)/integral_3)
print("mu",mu)
### CDFs

sigma_fail =4
if opts.as_test:
    # remember var is a VARIANCE
    if np.log(np.abs(integral_1b/integral_1)) > sigma_fail*np.sqrt(var_1)/integral_1:
        print(" FAIL AC")
        exit(1)
    if np.log(np.abs(integral_2/integral_1)) > sigma_fail*np.sqrt(var_1+var_2)/integral_1:
        print(" FAIL GMM ")
        exit(1)
    if np.log(np.abs(integral_3/integral_1)) > sigma_fail*np.sqrt(var_1+var_3)/integral_1:
        print(" FAIL AV ")
        exit(1)
        

if not(opts.save_plot):
    exit(0)

### get our posterior samples as a single array
arr_1 = np.empty((len(sampler._rvs["0"]), ndim))
arr_1b = np.empty((len(samplerAC._rvs["0"]), ndim))
arr_2 = np.empty((len(samplerEnsemble._rvs["0"]), ndim))
arr_3 = np.empty((len(samplerAV._rvs["0"]), ndim))
arr_4 = np.empty((len(samplerPortfolio._rvs["0"]), ndim))
for i in range(ndim):
    arr_1[:,i] = sampler._rvs[str(i)].flatten()
    arr_1b[:,i] = samplerAC._rvs[str(i)].flatten()
    arr_2[:,i] = samplerEnsemble._rvs[str(i)].flatten()
    arr_3[:,i] = samplerAV._rvs[str(i)].flatten()
    arr_4[:,i] = samplerPortfolio._rvs[str(i)].flatten()


if True:
    # NOTE: old mcsampler stores L, mcsamplerEnsemble stores lnL
    L = sampler._rvs["integrand"]
    p = sampler._rvs["joint_prior"]
    ps = sampler._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_1 = (L * p / ps)
    n_ess_1 = np.sum(weights_1)**2/np.sum(weights_1**2)
    print("default  n_eff, n_ess ", eff_samp_1,n_ess_1)

    L = samplerEnsemble._rvs["integrand"]
    if return_lnI:
        L = np.exp(L - np.max(L))
    p = samplerEnsemble._rvs["joint_prior"]
    ps = samplerEnsemble._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_2 = (L * p / ps)
    n_ess_2 = np.sum(weights_2)**2/np.sum(weights_2**2)
    print("AC  n_eff, n_ess ", eff_samp_2,n_ess_2)


    L = samplerAC._rvs["integrand"]
    p = samplerAC._rvs["joint_prior"]
    ps = samplerAC._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_1b = (L * p / ps)
    n_ess_1b = np.sum(weights_1b)**2/np.sum(weights_1b**2)
    print("GMM  n_eff, n_ess ", eff_samp_1b,n_ess_1b)

    lnL = samplerAV._rvs["log_integrand"]  
    lnp = samplerAV._rvs["log_joint_prior"]  
    lnps = samplerAV._rvs["log_joint_s_prior"]  
    ### compute weights of samples
    weights_3 = np.exp(lnL + lnp - lnps )
    n_ess_3 = np.sum(weights_3)**2/np.sum(weights_3**2)
    print("AV  n_eff, n_ess ", eff_samp_3,n_ess_3)


if save_fairdraws:
    npts_out_1 = int(n_ess_1)
    p = np.array(weights_1/np.sum(weights_1),dtype=np.float64)
    indx_save_1 = np.random.choice(np.arange(len(p)), size=npts_out_1, p=p)
    np.savetxt(save_fairdraw_prefix+"_1.dat", arr_1[indx_save_1],header=" x1 x2")

    npts_out_2 = int(n_ess_2)
    p = np.array(weights_2/np.sum(weights_2),dtype=np.float64)
    indx_save_2 = np.random.choice(np.arange(len(p)), size=npts_out_2, p=p)
    np.savetxt(save_fairdraw_prefix+"_2.dat", arr_2[indx_save_2],header=" x1 x2")

    npts_out_1b = int(n_ess_1b)
    p = np.array(weights_1b/np.sum(weights_1b),dtype=np.float64)
    indx_save_1b = np.random.choice(np.arange(len(p)), size=npts_out_1b, p=p)
    np.savetxt(save_fairdraw_prefix+"_1b.dat", arr_1b[indx_save_1b],header=" x1 x2")

    npts_out_3 = np.min([int(n_ess_3*0.8),5000])
    p = np.array(weights_3/np.sum(weights_3),dtype=np.float64)
    indx_save_3 = np.random.choice(np.arange(len(p)), size=npts_out_3, p=p)
    np.savetxt(save_fairdraw_prefix+"_3.dat", arr_3[indx_save_3],header=" x1 x2")


# Make copies, so they have the same orders relative to the sample values
weights_1_orig = np.array(weights_1)
weights_1b_orig = np.array(weights_1b)
weights_2_orig = np.array(weights_2)
weights_3_orig = np.array(weights_3)


colors = ["black", "red", "blue", "green", "orange"]

plt.figure(figsize=(10, 8))

for i in range(ndim):
    s = np.sqrt(cov[i][i])
    ### get sorted samples (for the current dimension)
    x_1 = arr_1[:,i][np.argsort(arr_1[:,i])]
    x_1b = arr_1b[:,i][np.argsort(arr_1b[:,i])]
    x_2 = arr_2[:,i][np.argsort(arr_2[:,i])]
    x_4 = arr_4[:,i][np.argsort(arr_4[:,i])]
    ### plot true cdf
    plt.plot(x_1, truncnorm.cdf(x_1, llim, rlim, mu[i], s), label="True",
            color=colors[i], linewidth=0.5)
    # NOTE: old mcsampler stores L, mcsamplerEnsemble stores lnL
    L = sampler._rvs["integrand"]
    p = sampler._rvs["joint_prior"]
    ps = sampler._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_1 = (L * p / ps)[np.argsort(arr_1[:,i])]
    L = samplerEnsemble._rvs["integrand"]
    if return_lnI:
        L = np.exp(L - np.max(L))
    p = samplerEnsemble._rvs["joint_prior"]
    ps = samplerEnsemble._rvs["joint_s_prior"]
    ### compute weights of samples
    weights_2 = (L * p / ps)[np.argsort(arr_2[:,i])]

    if "integrand" in samplerAC._rvs:
        L = samplerAC._rvs["integrand"]
        p = samplerAC._rvs["joint_prior"]
        ps = samplerAC._rvs["joint_s_prior"]
    else:
        L = np.exp(samplerAC._rvs["log_integrand"])
        p = np.exp(samplerAC._rvs["log_joint_prior"])
        ps = np.exp(samplerAC._rvs["log_joint_s_prior"])
    ### compute weights of samples
    weights_1b = (L * p / ps)[np.argsort(arr_1b[:,i])]

    weights_4 = (L * p / ps)[np.argsort(arr_4[:,i])]
    L = np.exp(samplerPortfolio._rvs["log_integrand"])
    p = np.exp(samplerPortfolio._rvs["log_joint_prior"])
    ps = np.exp(samplerPortfolio._rvs["log_joint_s_prior"])
    ### compute weights of samples
    weights_4 = (L * p / ps)[np.argsort(arr_4[:,i])]


    y_1 = np.cumsum(weights_1)
    y_1 /= y_1[-1] # normalize
    y_2 = np.cumsum(weights_2)
    y_2 /= y_2[-1] # normalize
    y_1b = np.cumsum(weights_1b)
    y_1b /= y_1b[-1]
    y_4 = np.cumsum(weights_4)
    y_4 /= y_4[-1] # normalize
    ### plot recovered cdf
    plt.plot(x_1, y_1, "--", label="mcsampler", color=colors[i],
            linewidth=2)
    plt.plot(x_2, y_2, ":", label="mcsamplerEnsemble",
            color=colors[i], linewidth=2)
    plt.plot(x_1b, y_1b, ".", label="mcsamplerAC",
            color=colors[i], linewidth=1)
    plt.plot(x_4, y_4, ":", label="mcsamplerPortfolio",
            color=colors[i], linewidth=2)

#plt.legend()

fname = "cdf.pdf"

print("Saving CDF figure as " + fname + "...")
plt.xlim(-width/2,width/2)
plt.xlabel(r'$x$')
plt.ylabel(r'$P(<x)$')
plt.legend()
plt.savefig(fname)


# TEST OUTPUT



#! /usr/bin/env python
#
# python convergence_test_lnL.py --composite-file all.net --flag-tides-in-composite --d-eff 4
# for i in rest*/run*/all.net; do echo $i; python ~/convergence_test_lnL.py --composite-file $i --d-eff 8 --eccentricity --meanPerAno; done


import numpy as np
import argparse
import scipy.stats
from scipy.stats import gaussian_kde, chi2
import numpy.linalg as la
import sys

import RIFT.lalsimutils as lalsimutils
import RIFT.misc.samples_utils as samples_utils
from RIFT.misc.samples_utils import add_field
from RIFT.misc.samples_utils import add_field, extract_combination_from_LI, standard_expand_samples

parser = argparse.ArgumentParser()
parser.add_argument("--composite-file", action='append', help="Samples used in convergence test")
parser.add_argument("--composite-file-has-labels",action='store_true',help="Assume header for composite file")
parser.add_argument("--flag-tides-in-composite",action='store_true',help='Required, if you want to parse files with tidal parameters')
parser.add_argument("--flag-eos-index-in-composite",action='store_true',help='Required, if you want to parse files with EOS index in composite (and tides)')
parser.add_argument("--lnL-cut",default=15,type=float)
parser.add_argument("--sigma-cut",default=0.4,type=float)
parser.add_argument("--d-eff",default=3,type=int,help="Effective dimension, used to assess if we have enough nearby samples close to the peak. ")
parser.add_argument("--n-close-min",default=1,type=int,help="Minimum number of samples close to the peak, *per effective dimension* (within 1)")
parser.add_argument("--n-cut-min",default=2000,type=int,help="Minimum number of samples passing the likelihood cut")
parser.add_argument("--eccentricity", action="store_true", help="Read sample files in format including eccentricity")
parser.add_argument("--meanPerAno", action="store_true", help="Read sample files in format including meanPerAno - assumes eccentricity also present")
parser.add_argument("--posterior-file",default=[], action='append', help="Samples used in convergence test")
parser.add_argument("--parameter", action='append', help="Parameters used in convergence test")
parser.add_argument("--threshold",default=0.01,type=float,  help="Manual threshold for the test being performed. (If not specified, the success condition is determined by default for that diagnostic, based on the samples size and properties).  Try 0.01")
parser.add_argument("--tests", action='append',default=[],  help="What tests to perform")
parser.add_argument("--test-output",  help="Filename to return output. Result is a scalar >=0 and ideally <=1.  Closer to 0 should be good. Second column is the diagnostic, first column is 0 or 1 (success or failure)")
parser.add_argument("--always-succeed",action='store_true',help="Test output is always success.  Use for plotting convergence diagnostics so jobs insured to run for many iterations.")
parser.add_argument("--iteration-threshold",default=0,type=int,help="Test is applied if iteration >= iteration-threshold. Default is 0")
parser.add_argument("--iteration",default=0,type=int,help="Current reported iteration. Default is 0.")
parser.add_argument("--write-file-on-success",type=str,default="INTRINSIC_CONVERGED",help="Produces an (empty) file with this name if the convergence tests passes.  Note you should pass the FULL PATH to this file if you want it to occur in the run directory for example")
opts=  parser.parse_args()

if len(opts.composite_file)<1:
    print(" Need at least one composite file ")
    sys.exit(1)

if opts.iteration < opts.iteration_threshold:
    sys.exit(0)

field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","lnL", "sigmaOverL", "ntot", "neff")
if opts.flag_tides_in_composite:
    if opts.flag_eos_index_in_composite:
        print(" Reading composite file, assumingtide/eos-index-based format ")
        field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","lambda1", "lambda2", "eos_indx","lnL", "sigmaOverL", "ntot", "neff")
    else:
        print(" Reading composite file, assuming tide-based format ")
        field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","lambda1", "lambda2", "lnL", "sigmaOverL", "ntot", "neff")
if opts.eccentricity:
    print(" Reading composite file, assuming eccentricity-based format ")
    if opts.meanPerAno:
        print(" Reading composite file, assuming mpa-based format ")
        field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","eccentricity", "meanPerAno", "lnL", "sigmaOverL", "ntot", "neff")
    else:
        field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","eccentricity", "lnL", "sigmaOverL", "ntot", "neff")
field_formats = [np.float32 for x in field_names]
composite_dtype = [ (x,float) for x in field_names] #np.dtype(names=field_names ,formats=field_formats)

    
samples_list = []
composite_list = []
for fname in opts.posterior_file:
    samples = np.genfromtxt(fname, names=True)
    samples = stan
    samples = standard_expand_samples(samples)
    for name in samples.dtype.names:
        if name in lalsimutils.periodic_params:
            samples[name] = np.mod(samples[name], lalsimutils.periodic_params[name])
    samples_list.append(samples)
# Composite file parsing: from composite list
for fname in opts.composite_file:
    if not(opts.composite_file_has_labels):
        samples = np.loadtxt(fname,dtype=composite_dtype)  # Names are not always available
    else:
        samples = np.genfromtxt(fname,names=True)
        samples = rfn.rename_fields(samples, {'sigmalnL': 'sigmaOverL', 'sigma_lnL': 'sigmaOverL'})   # standardize names, some drift in labels
    # enforce periodicity
    for name in samples.dtype.names:
        if name in lalsimutils.periodic_params:
            samples[name] = np.mod(samples[name], lalsimutils.periodic_params[name])
    if 'lnL' in samples.dtype.names:
        samples = samples[ ~np.isnan(samples["lnL"])] # remove nan likelihoods -- they can creep in with poor settings/overflows
    name_ref = samples.dtype.names[0]

    if 'm1' in samples.dtype.names:
        samples=add_field(samples,[('mtotal',float)]); samples["mtotal"]= samples["m1"]+samples["m2"]; 
        samples=add_field(samples,[('q',float)]); samples["q"]= samples["m2"]/samples["m1"]; 
        samples=add_field(samples,[('mc',float)]); samples["mc"] = lalsimutils.mchirp(samples["m1"], samples["m2"])
        samples=add_field(samples,[('eta',float)]); samples["eta"] = lalsimutils.symRatio(samples["m1"], samples["m2"])
        samples=add_field(samples,[('chi_eff',float)]); samples["chi_eff"]= (samples["m1"]*samples["a1z"]+samples["m2"]*samples["a2z"])/(samples["mtotal"]); 
        chi1_perp = np.sqrt(samples['a1x']*samples["a1x"] + samples['a1y']**2)
        chi2_perp = np.sqrt(samples['a2x']**2 + samples['a2y']**2)
        samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
        samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp

        phi1 = np.arctan2(samples['a1x'], samples['a1y']);
        phi2 = np.arctan2(samples['a2x'], samples['a2y']);
        samples = add_field(samples, [('phi1',float), ('phi2',float), ('phi12',float)])
        samples['phi1'] = phi1
        samples['phi2'] = phi2
        samples['phi12'] = phi2 - phi1

        if ('lambda1' in samples.dtype.names):
            Lt,dLt = lalsimutils.tidal_lambda_tilde(samples['m1'], samples['m2'],  samples['lambda1'], samples['lambda2'])
            samples= add_field(samples, [('LambdaTilde',float), ('DeltaLambdaTilde',float),('lambdat',float),('dlambdat',float)])
            samples['LambdaTilde'] = samples['lambdat']= Lt
            samples['DeltaLambdaTilde'] = samples['dlambdat']= dLt

    
    if opts.sigma_cut >0:
        npts = len(np.atleast_1d(samples[name_ref]))
        # strip NAN
        sigma_vals = samples["sigmaOverL"]
        good_sigma = sigma_vals < opts.sigma_cut
        npts_out = np.sum(good_sigma)
        if npts_out < npts:
            new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
            for name in samples.dtype.names:
                new_samples[name] = samples[name][good_sigma]
            samples = new_samples
    if opts.lnL_cut and 'lnL' in samples.dtype.names:
        npts = len(np.atleast_1d(samples[name_ref]))
        # strip NAN
        lnL_vals = samples["lnL"]
        not_nan = np.logical_not(np.isnan(lnL_vals))
        npts_out = np.sum(not_nan)
        if npts_out < npts:
            new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
            for name in samples.dtype.names:
                new_samples[name] = samples[name][not_nan]
            samples = new_samples
        
        # apply cutoff
        indx_ok =np.arange(npts)
        lnL_max = np.max(samples["lnL"])
        indx_ok = samples["lnL"]>lnL_max  -opts.lnL_cut
        npts_out = np.sum(indx_ok)
        new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
        for name in samples.dtype.names:
            new_samples[name] = samples[name][indx_ok]
        samples = new_samples
    print(" Reading ", len(samples), " from {} : lnLmax = {}".format(fname, lnL_max))
    composite_list.append(samples)


#print(samples_list, composite_list)
composite_preferred = composite_list[0]
thresh_max = -opts.d_eff
lnLmax =  np.max(composite_preferred['lnL'])  # assume anything passing is good
n_close = np.sum(composite_preferred['lnL'] > lnLmax + thresh_max)  # typical dimension
thresh_levels = np.array([-1,-2,-3,-4,-5, -6, thresh_max])
expected_cdf = scipy.stats.chi2.cdf( -  2*thresh_levels,df=opts.d_eff)
expected_cdf *= 1./expected_cdf[-1]
n_thresh = np.array([np.sum( composite_preferred['lnL'] > lnLmax  + offset) for offset in thresh_levels])
print(" Number of nearby samples " , n_thresh)
print(" Fraction of nearby samples ", n_thresh/n_close, expected_cdf)
print(" Ratio ", n_thresh/n_close / expected_cdf)

if len(composite_preferred) < opts.n_cut_min:
    print("  Failure: too few points contribute to estimating posterior , absolute. ")
    sys.exit(1)
    
if n_thresh[0] < opts.n_close_min*opts.d_eff:
    print("  Failure: too few points near peak , absolute. FIXME MODEL ")
    sys.exit(1)
if (n_thresh/n_close/expected_cdf)[0] < 1e-2:
    print("  Failure: too few points near peak , vs expected ")
    sys.exit(1)
sys.exit(0)    



from factored_likelihood import *
from matplotlib import pylab as plt

m1 = 5*lal.LAL_MSUN_SI
m2 = 3*lal.LAL_MSUN_SI
ampO = 0 # sets which modes to include in the physical signal
Lmax = 2  # sets which modes to include in the output
distMpcReference = 100
distMpcFiducial = 25

Psig = ChooseWaveformParams(fmin = 30., radec=False, incl=0.0, theta=0, phi=0,
         m1=m1,m2=m2,
         ampO=ampO,
        detector='H1', dist=distMpcFiducial*1.e6*lal.LAL_PC_SI)
df = findDeltaF(Psig)
Psig.deltaF = df
Psig.print_params()
data = lsu.non_herm_hoff(Psig)
psd = lal.LIGOIPsd
IP = ComplexOverlap(30, 2048, data.deltaF, psd, True, True)



# Generate and plot hlm(f) and the weights we would use in an inner product
Psig.dist = distMpcReference*1.e6*lal.LAL_PC_SI  # Fiducial distance
hlms = hlmoff(Psig,2)

plt.figure(1)
fvals = np.linspace(-IP.fNyq, IP.fNyq, IP.wvlen)
plt.plot(fvals, 1/np.sqrt(IP.longweights))
plt.xlim(-500,500)
plt.ylim(0, 1e-21)
plt.figure(2)
plt.plot(np.log10(np.abs(fvals)), np.log10(1/np.sqrt(IP.longweights)))
for m in range(-2,3):
    plt.figure(1)
    hlm = lalsim.SphHarmFrequencySeriesGetMode(hlms, 2, m)
    plt.plot(fvals,np.sqrt(np.abs(fvals))*np.abs(hlm.data.data),label=m)
    plt.figure(2)
    plt.plot(np.log10(np.abs(fvals)),np.log10(np.sqrt(np.abs(fvals))*np.abs(hlm.data.data)),label=m)


plt.show()



#! /usr/bin/env python
#
# OBJECTIVE
#    unit tests for RIFT outcomes, to make sure run completes successfully (run quality test)

from asimov.testing import AsimovTest
import os, sys,glob
import numpy as np


class TestRIFTConvergence(AsimovTest):
    def test_final_internal_convergence(self):
        """
        the rundir/iter*test/logs/*.out have convergence information : make sure we achieve our target (<0.02)
        """
        print('RIFT check for internal convergence test NOT IMPLEMENTED')
        for event in self.events:
            for production in event.productions:
                if production.pipeline.name.lower() == "rift":
                    with self.subTest("convergence", production=production.name):
                        rundir = production.rundir
#                        print('RIFT convergence test for ', rundir, "NOT IMPLEMENTED")
#                        repo = event.event_object.repository.directory


class TestRIFTCalmargSampleSize(AsimovTest):
    def test_calmarg_sample_size(self):
        """
        this will count the size of the number of distinct samples in the final file, and require it is more than 1000
        """
        for event in self.events:
            for production in event.productions:
                if production.pipeline.name.lower() == "rift":
                    rundir = production.rundir
                    print(rundir)
                    target_file = rundir + "/reweighted_posterior_samples.dat"
                    if os.path.exists(target_file):
                        with self.subTest(event=event, production=production.name):
                            dat = np.loadtxt(target_file,skiprows=1) # skip header row
                            print(event, production.name, len(dat))  # verbose, most common failure mode, can easily eyeball
                            self.assertFalse(  len(dat) <  1000  )



class TestRIFTFinalCIPneff(AsimovTest):
    def test_final_cip_neff(self):
        """
        this will look at the last CIP directory and assess the n_eff reported in each worker file.  None should be small, if the file is nonempty
        """
        for event in self.events:
            for production in event.productions:
                if production.pipeline.name.lower() == "rift":
                    rundir = production.rundir
                    # completion check
                    target_file = rundir + "/extrinsic_posterior_samples.dat"
                    if not(os.path.exists(target_file)):
                        continue

                    dir_cip_last = glob.glob(rundir +  "/iteration_*_cip")  # usually iteration_6_cip
                    dir_cip_last.sort()
                    if dir_cip_last is None or dir_cip_last is []:
                        print("   RIFT CIP n_eff : -should not happen-, skipping cip test ")
                        continue
                    dir_cip_last = dir_cip_last[-2] # second to last
                    with self.subTest(event=event, production=production.name):
                        cip_out_annotate  = glob.glob(dir_cip_last + "/overlap-grid-*_withpriorchange+annotation.dat")
#                        print(dir_cip_last, cip_out_annotate)
                        if len(cip_out_annotate) < 1:
                            print("   RIFT CIP n_eff stage 2 : -should not happen-, skipping cip test ")
                        n_eff_list = []
                        for name in cip_out_annotate:
                            dat  = np.genfromtxt(name, names=True)
                            n_eff_list.append(dat['neff'])
                        print(event, production.name, np.min(n_eff_list), 1000/len(n_eff_list) )
                        self.assertFalse( any(np.array(n_eff_list) < 1000/len(n_eff_list)) )   # attempt to require the sum of n_eff is at least 1000, distributed evenly over all of them

class TestRIFTMarginalizedLikelihoods(AsimovTest):
    def test_lnLmarg_final(self):
        """
        this will look at the 'all.net' file in the final run, and assess if we have too few high-likleihood points, or other pathologies such as large marginal likleihood error
        Note that to read this file you will need to know if it is eccentric, including matter,  or not, DEFAULT FILE READING
        """

        for event in self.events:
            for production in event.productions:
                if production.pipeline.name.lower() == "rift":
                    rundir = production.rundir
                    # completion check
                    target_file = rundir + "/extrinsic_posterior_samples.dat"
                    if not(os.path.exists(target_file)):
                        continue
                    # test on marginalized likelihood
                    all_net_name = production.pipeline.collect_assets()['lnL_marg']
                    # check file has nonzero size
                    if os.stat(all_net_name).st_size ==0:
                        continue
                    dat = np.loadtxt(all_net_name)  # lnL col is -3, always
                    if len(dat) < 1:
                        continue 
                    lnL_col = -4
                    sigma_col = -3
                    # sort by lnL
                    indx_sort = np.argsort( dat[:,lnL_col])
                    dat = dat[indx_sort]
                    # syntactic sugar
                    lnL_vals = dat[:,lnL_col]
                    sigma_vals = dat[:,sigma_col]
                    n_lines = len(dat)
                    with self.subTest(event=event, production=production.name):
                        # Check file size
                        self.assertFalse(len(lnL_vals)< 5000) # too few lines
                    with self.subTest(event=event, production=production.name):
                        # check top 10% of file has reasonable sigma_vals. Use MEAN
                        n_test = int(n_lines*0.1)
                        self.assertFalse( np.mean(sigma_vals[:n_test]) > 0.4)

                    with self.subTest(event=event, production=production.name):
                        # check that we have enough points within 10 of the peak (typical model dimension constraint).  Argue we need at least 1000
                        self.assertFalse( np.sum( np.max(lnL_vals) - lnL_vals < 10 ) < 1000)

                        
if __name__ == '__main__':
    import unittest
    unittest.main()                    



#!/usr/bin/env python
#  based on 'fast_tutorial.py' from https://git.ligo.org/lscsoft/bilby/-/blob/master/examples/gw_examples/injection_examples/fast_tutorial.py
#  and https://git.ligo.org/lscsoft/bilby/-/blob/master/examples/gw_examples/injection_examples/plot_time_domain_data.py
#
# INTENT:
#    tests rift_source, for use in calmarg
# EXAMPLE
#     python compare_in_ifo.py --inj ../../mdc.xml.gz --rift-wf --instrument H1

import bilby
import lal
from matplotlib import pyplot as plt

import numpy as np
#from gwpy.timeseries import TimeSeries
#import glob
from RIFT.calmarg.rift_source import RIFT_lal_binary_black_hole
import RIFT.lalsimutils as lalsimutils

import argparse
parser = argparse.ArgumentParser()
parser.add_argument("--inj",default=None)
parser.add_argument("--approx", default="IMRPhenomPv2",help="Use H1, L1,V1")
parser.add_argument("--instrument", default=[],action='append',help="Use H1, L1,V1")
parser.add_argument("--rift-wf", action='store_true',help="Use rift wf generator")
parser.add_argument("--guess-offset",default=None, type=float,help="Hack: sign flip the polarization change")
parser.add_argument("--guess-sign-flip", action='store_true',help="Hack: sign flip the polarization change")
parser.add_argument("--seglen",default=4,type=int)
parser.add_argument("--fmin",default=25,type=float)
opts=  parser.parse_args()


ifo_list =['H', 'L', 'V']
my_data = {}
event_time = 1000000010.000
start_time = event_time-2
end_time =event_time +2


current_ifo=opts.instrument




# Set the duration and samplingg frequency of the data segment that we're
# going to inject the signal into
duration = float(opts.seglen)
sampling_frequency = 2048.0
minimum_frequency = opts.fmin   # avoid 2s duration before trigger time!


if opts.inj:
    P = lalsimutils.xml_to_ChooseWaveformParams_array(opts.inj)[0]
else:
    P = lalsimutils.ChooseWaveformParams()
    P.m1 = 25*lal.MSUN_SI
    P.m2 = 20*lal.MSUN_SI
    P.theta= 1.3
    P.phi = -1.57
    P.incl = 1.57
    P.psi = 0.5
    P.phiref = 1
    P.dist = 1*lal.PC_SI*1e6
P.radec=True # real injection
P.deltaF = 1./duration
P.deltaT = 1./sampling_frequency
P.fmin= minimum_frequency
    


# Specify the output directory and the name of the simulation.
outdir = "outdir_plots_"+''.join(current_ifo)
label = "from_file"
bilby.core.utils.setup_logger(outdir=outdir, label=label)

# Set up a random seed for result reproducibility.  This is optional!
bilby.core.utils.random.seed(88170235)

# We are going to inject a binary black hole waveform.  We first establish a
# dictionary of parameters that includes all of the different waveform
# parameters, including masses of the two black holes (mass_1, mass_2),
# spins of both black holes (a, tilt, phi), etc.
# SHOULD read parameters from mdc.xml.gz file
injection_parameters = dict(
    mass_1=P.m1/lal.MSUN_SI,
    mass_2=P.m2/lal.MSUN_SI,
    a_1=0.0,
    a_2=0.0,
    tilt_1=0,
    tilt_2=1.0,
    phi_12=0,
    phi_jl=0.0,
    luminosity_distance=P.dist/(lal.PC_SI*1e6), # force same distance
    theta_jn=np.pi/2,
    psi=P.psi,
    phase=P.phiref,
    geocent_time=1000000010.0001,
    ra=P.phi,
    dec=P.theta,
)
print(injection_parameters)
P.print_params()   # DOES NOT MATCH UP
if opts.guess_sign_flip:
    injection_parameters['psi'] = -injection_parameters['psi']
if opts.guess_offset:
    injection_parameters['psi'] += opts.guess_offset*np.pi # scale
    

# Set up interferometers.  In this case we'll use two interferometers
# (LIGO-Hanford (H1), LIGO-Livingston (L1). These default to their design
# sensitivity
ifos = bilby.gw.detector.InterferometerList(current_ifo)
ifos.set_strain_data_from_power_spectral_densities(
    sampling_frequency=sampling_frequency,
    duration=duration,
    start_time=injection_parameters["geocent_time"] - 2,
)
# set to zero
for ifo in ifos:
    ifo.frequency_domain_strain = np.zeros(len(ifo.frequency_domain_strain),dtype=np.complex128)



# Fixed arguments passed into the source model
waveform_arguments = dict(
    waveform_approximant=opts.approx,
    reference_frequency=50.0,
    minimum_frequency=minimum_frequency,
    maximum_frequency=1024,
    pn_amplitude_order=0
)

source_model=bilby.gw.source.lal_binary_black_hole
conversion_function = bilby.gw.conversion.convert_to_lal_binary_black_hole_parameters
if opts.rift_wf:
    source_model=RIFT_lal_binary_black_hole
    def clean_converter(*args,**kwargs):
        out = bilby.gw.conversion.generate_all_bbh_parameters(*args, **kwargs)
        if isinstance(out,dict):
            return (out, [])   # hack adding second return value, if it is not working correctlyx
        else:
            return out
    conversion_function =    clean_converter


# Create the waveform_generator using a LAL BinaryBlackHole source function
waveform_generator = bilby.gw.WaveformGenerator(
    duration=duration,
    sampling_frequency=sampling_frequency,
    frequency_domain_source_model=source_model,
    parameter_conversion=conversion_function,
    waveform_arguments=waveform_arguments,
)

# inject into each interferometer. Remember it is frequency-domain data
hT_dict={}
hf_signal = waveform_generator.frequency_domain_strain(injection_parameters)
for ifo in ifos:
    ifo.minimum_frequency=minimum_frequency
    print(ifo)
    ifo.inject_signal(injection_polarizations=hf_signal, parameters=injection_parameters)
    P.detector=ifo.name
    hT_dict[ifo.name] = lalsimutils.hoft(P)
    
# extract the data using self.time_domain_strain, based on bilby.gw.detector.inferferometer
extra_name =''
if opts.guess_sign_flip:
    extra_name += '-flipped'
for ifo in ifos:
    tvals = ifo.strain_data.time_array
    hvals = ifo.strain_data.time_domain_strain
    plt.plot(tvals - injection_parameters['geocent_time'],hvals)
    tvals_R =lalsimutils.evaluate_tvals( hT_dict[ifo.name])
    plt.xlim(-0.5,0.1)
    plt.plot(tvals_R -  injection_parameters['geocent_time'], hT_dict[ifo.name].data.data)
    plt.savefig("my_compare_{}{}.png".format(ifo.name,extra_name))
    plt.clf()

exit(0)
### additional framework
# Initialise the likelihood by passing in the interferometer data (ifos) and
# the waveform generator
likelihood = bilby.gw.GravitationalWaveTransient(
    interferometers=ifos, waveform_generator=waveform_generator
)


# using transient code, to get raw waveform data
#waveform_polarizations = \
#    likelihood.waveform_generator.frequency_domain_strain(self.parameters)





from  lalsimutils import *
from marginalized_likelihood import *

P = ChooseWaveformParams()

psd=lalsim.SimNoisePSDiLIGOModel

Lmax = 5

fSample = 4096.


hlms = lalsim.SimInspiralChooseTDModes(P.phiref, P.deltaT, P.m1, P.m2,
                                      # P.s1x, P.s1y, P.s1z, P.s2x, P.s2y, P.s2z, 
                                       P.fmin, P.fref, P.dist,
                                      # P.incl, 
                                       P.lambda1, P.lambda2, P.waveFlags, P.nonGRparams,
                                       P.ampO, P.phaseO, Lmax, P.approx)


crossTerms = ComputeModeCrossTermIP(hlms, psd, P.fmin, fSample/2.)

print(crossTerms)




from factored_likelihood import *
from matplotlib import pylab as plt

m1 = 5*lal.LAL_MSUN_SI
m2 = 3*lal.LAL_MSUN_SI
ampO = 0 # sets which modes to include in the physical signal
Lmax = 2  # sets which modes to include in the output
distMpcReference = 100
distMpcFiducial = 25

# optimally oriented source in one detector
Psig = ChooseWaveformParams(fmin = 30., radec=False, incl=0.0, theta=0, phi=0,
         m1=m1,m2=m2,
         ampO=ampO,
        detector='H1', dist=distMpcFiducial*1.e6*lal.LAL_PC_SI)
df = findDeltaF(Psig)
Psig.deltaF = df
Psig.print_params()
data = lsu.non_herm_hoff(Psig)
psd = lal.LIGOIPsd
IP = ComplexOverlap(30, 2048, data.deltaF, psd, True, True)

rhoData = IP.norm(data)  # Amplitude of the actual real signal in the data
print("Amplitude in one IFO : ", rhoData)

Psig.dist = distMpcReference*1.e6*lal.LAL_PC_SI  # Fiducial distance
hlms = hlmoff(Psig,2)
hlm = lalsim.SphHarmFrequencySeriesGetMode(hlms, 2, 2)
rho, rhoTS, rhoIdx, rhoPhase = IP.ip(hlm, data)


# we do have wraparound; we need to force it back around
Tmax = rhoTS.deltaT* len(rhoTS.data.data)
t =map( lambda x:  x if x< Tmax/2 else x-Tmax,  rhoTS.deltaT* np.arange(len(rhoTS.data.data)))

fvals = np.linspace(-2048, 2048, len(t))

plt.figure(1)
plt.plot(t, np.sqrt(np.abs(rhoTS.data.data)*distMpcReference/distMpcFiducial)*np.sqrt(5./(4.*lal.LAL_PI)))  # rescale to match
plt.plot(t, np.abs(rhoData)*np.ones(len(t)))
plt.xlim(-0.05,0.05)
plt.show()

# plt.figure(2)
# plt.plot(fvals, np.abs(hlm.data.data))
# plt.plot(fvals, np.abs(data.data.data))
# plt.show()




#! /usr/bin/env python
#
# GOAL
#   - read latest lnLmarg file from target working directory
#   - assess its distribution (modulo errors)
#       * Warning if lots of high-errorr points (esp at top
#       * Plot distribution of lnL, probbly warn if long tail but peak lnL is high
#   - probably provide option to run as a convergence test!
#   - WARNING; 'composite' includes the PUFFBALL points, and so it's not a fair test!


import RIFT.lalsimutils as lalsimutils
import RIFT.misc.bounded_kde as bounded_kde
from  RIFT.misc.samples_utils import  add_field

import scipy.stats
import glob

import lal
import numpy as np
import argparse


try:
    import matplotlib
#    print(" Matplotlib backend ", matplotlib.get_backend())
    if matplotlib.get_backend() == 'agg':
        fig_extension = '.png'
        bNoInteractivePlots=True
    else:
        matplotlib.use('agg')
        fig_extension = '.png'
        bNoInteractivePlots =True
    from matplotlib import pyplot as plt
    bNoPlots=False
except:
    print(" Error setting backend")

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import matplotlib.lines as mlines
import corner

dpi_base=200
legend_font_base=16
rc_params = {'backend': 'ps',
             'axes.labelsize': 11,
             'axes.titlesize': 10,
             'font.size': 11,
             'legend.fontsize': legend_font_base,
             'xtick.labelsize': 11,
             'ytick.labelsize': 11,
             #'text.usetex': True,
             'font.family': 'Times New Roman'}#,
             #'font.sans-serif': ['Bitstream Vera Sans']}#,
plt.rcParams.update(rc_params)




parser = argparse.ArgumentParser()
parser.add_argument("--run-directory",help="run directory (preferred method).  Will auto-identify file")
parser.add_argument("--test",action='store_true')
parser.add_argument("--assume-dof",default=None,type=int,help="Number of DOF to assume for chisquared comparison")
parser.add_argument("--composite-file",help="Specific composite file to use (single), assumes fair draw")
parser.add_argument("--flag-tides-in-composite",action='store_true',help='Required, if you want to parse files with tidal parameters')
parser.add_argument("--eccentricity", action="store_true", help="Read sample files in format including eccentricity")
parser.add_argument("--sigma-cut",default=0.4,type=float)
parser.add_argument("--chi-max",default=1,type=float)
opts=  parser.parse_args()


# Identify files
if opts.run_directory:
    mynames = list(glob.glob(opts.run_directory+"/*.composite"))
    mynames.sort(key=lambda f: int(filter(str.isdigit, f)))
    print(" Composite files : ", mynames)
    opts.composite_file = mynames[-1]


# Import
composite_list = []
composite_full_list = []
field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","lnL", "sigmaOverL", "ntot", "neff")
if opts.flag_tides_in_composite:
    print(" Reading composite file, assuming tide-based format ")
    field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","lambda1", "lambda2", "lnL", "sigmaOverL", "ntot", "neff")
if opts.eccentricity:
    print(" Reading composite file, assuming eccentricity-based format ")
    field_names=("indx","m1", "m2",  "a1x", "a1y", "a1z", "a2x", "a2y", "a2z","eccentricity", "lnL", "sigmaOverL", "ntot", "neff")
field_formats = [np.float32 for x in field_names]
composite_dtype = [ (x,float) for x in field_names] #np.dtype(names=field_names ,formats=field_formats)
# Load composite file
if opts.composite_file:
    fname = opts.composite_file
    print(" Loading ... ", fname)
    samples = np.loadtxt(fname,dtype=composite_dtype)  # Names are not always available
    # DO NOT CUT if we are testing!
    if not(opts.test) and opts.sigma_cut >0:
        npts = len(samples["m1"])
        # strip NAN
        sigma_vals = samples["sigmaOverL"]
        good_sigma = sigma_vals < opts.sigma_cut
        npts_out = np.sum(good_sigma)
        if npts_out < npts:
            new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
            for name in samples.dtype.names:
                new_samples[name] = samples[name][good_sigma]
            samples = new_samples

#    samples = np.recarray(samples.T,names=field_names,dtype=field_formats) #,formats=field_formats)
    # If no record names
    # Add mtotal, q, 
    samples=add_field(samples,[('mtotal',float)]); samples["mtotal"]= samples["m1"]+samples["m2"]; 
    samples=add_field(samples,[('q',float)]); samples["q"]= samples["m2"]/samples["m1"]; 
    samples=add_field(samples,[('mc',float)]); samples["mc"] = lalsimutils.mchirp(samples["m1"], samples["m2"])
    samples=add_field(samples,[('eta',float)]); samples["eta"] = lalsimutils.symRatio(samples["m1"], samples["m2"])
    samples=add_field(samples,[('chi_eff',float)]); samples["chi_eff"]= (samples["m1"]*samples["a1z"]+samples["m2"]*samples["a2z"])/(samples["mtotal"]); 
    chi1_perp = np.sqrt(samples['a1x']*samples["a1x"] + samples['a1y']**2)
    chi2_perp = np.sqrt(samples['a2x']**2 + samples['a2y']**2)
    samples = add_field(samples, [('chi1_perp',float)]); samples['chi1_perp'] = chi1_perp
    samples = add_field(samples, [('chi2_perp',float)]); samples['chi2_perp'] = chi2_perp

    if ('lambda1' in samples.dtype.names):
        Lt,dLt = lalsimutils.tidal_lambda_tilde(samples['m1'], samples['m2'],  samples['lambda1'], samples['lambda2'])
        samples= add_field(samples, [('LambdaTilde',float), ('DeltaLambdaTilde',float),('lambdat',float),('dlambdat',float)])
        samples['LambdaTilde'] = samples['lambdat']= Lt
        samples['DeltaLambdaTilde'] = samples['dlambdat']= dLt

    samples_orig = samples

    npts = len(samples["m1"])
        # strip NAN
    lnL_vals = samples["lnL"]
    not_nan = np.logical_not(np.isnan(lnL_vals))
    npts_out = np.sum(not_nan)
    if npts_out < npts:
        new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
        for name in samples.dtype.names:
            new_samples[name] = samples[name][not_nan]
        samples = new_samples
        

    print(" Loaded samples from ", fname , len(samples["m1"]))
    if True:
        # impose Kerr limit
        npts = len(samples["m1"])
        indx_ok =np.arange(npts)
        chi1_squared = samples["chi1_perp"]**2 + samples["a1z"]**2
        chi2_squared = samples["chi2_perp"]**2 + samples["a2z"]**2
        indx_ok = np.logical_and(chi1_squared < opts.chi_max ,chi2_squared < opts.chi_max)
        npts_out = np.sum(indx_ok)
        if npts_out < npts:
            print(" Ok systems ", npts_out)
            new_samples = np.recarray( (npts_out,), dtype=samples.dtype)
            for name in samples.dtype.names:
                new_samples[name] = samples[name][indx_ok]
            samples = new_samples
            print(" Stripped samples  from ", fname , len(samples["m1"]))


    composite_list.append(samples)
    composite_full_list.append(samples_orig)

#    continue


lnLmax = np.max(samples["lnL"])

###
### WARNINGS
###

if opts.test:
    # Test 0: Lots of high variance points within top 10%
    npts = len(samples['lnL'])
    npts_test = int(0.1*npts)+1
    vals = np.array(samples["lnL"])
    lnL_crit = vals[-npts_test]
    indx_top = samples["lnL"]>lnL_crit
    indx_top_bad = np.logical_and(samples["lnL"] > lnL_crit, samples["sigmaOverL"] > opts.sigma_cut) # cannot pass if we remove!
    if np.sum(indx_top_bad)/ np.sum(indx_top) >  0.01:
        print(" WARNING: High error points are more than 1% of the final sample; consider increasing n_eff or n_max for ILE ")


###
### PLOTS
###

# Generally we do NOT want to make these if we are in test mode, since we want to reject points

# Step 1: lnL cdf
#   - overlay expected DOF
vals = np.array(samples["lnL"])
n_dof =2
if np.any(np.abs(samples["a1z"])>0):
    n_dof +=2
if np.any(np.abs(samples["a1x"])>0):
    n_dof +=4
vals.sort()
plt.plot( vals, np.arange(len(vals))/(1.*len(vals)))
plt.plot( vals, 1-scipy.stats.chi2.cdf( (lnLmax-vals)*2,df=n_dof))
plt.xlabel(r'$\ln L$')
plt.ylabel(r'$P(<\ln L)$')
plt.savefig("cdf_lnL.png")



import sys
import math
#import bisect
from collections import defaultdict
from types import ModuleType

import numpy
np=numpy #import numpy as np
from scipy import integrate, interpolate, special
import itertools
import functools

xpy_default= numpy


class MCSamplerGeneric(object):
    """
    Class to define a set of parameter names, limits, and probability densities.
    """


    @staticmethod
    def match_params_from_args(args, params):
        """
        Given two unordered sets of parameters, one a set of all "basic" elements (strings) possible, and one a set of elements both "basic" strings and "combined" (basic strings in tuples), determine whether the sets are equivalent if no basic element is repeated.
        e.g. set A ?= set B
        ("a", "b", "c") ?= ("a", "b", "c") ==> True
        (("a", "b", "c")) ?= ("a", "b", "c") ==> True
        (("a", "b"), "d")) ?= ("a", "b", "c") ==> False  # basic element 'd' not in set B
        (("a", "b"), "d")) ?= ("a", "b", "d", "c") ==> False  # not all elements in set B represented in set A
        """
        not_common = set(args) ^ set(params)
        if len(not_common) == 0:
            # All params match
            return True
        if all([not isinstance(i, tuple) for i in not_common]):
            # The only way this is possible is if there are
            # no extraneous params in args
            return False

        to_match, against = [i for i in not_common if not isinstance(i, tuple)], [i for i in not_common if isinstance(i, tuple)]

        matched = []
        import itertools
        for i in range(2, max(list(map(len, against)))+1):
            matched.extend([t for t in itertools.permutations(to_match, i) if t in against])
        return (set(matched) ^ set(against)) == set()




    def __init__(self,**kwargs):

        # Total number of samples drawn
        self.ntotal = 0
        # Parameter names
        self.params = set()
        self.params_ordered = []  # keep them in order. Important to break likelihood function need for names
        # If the pdfs aren't normalized, this will hold the normalization 
        # Cache for the sampling points
        self._rvs = {}
        # parameter -> cdf^{-1} function object
        # params for left and right limits
        self.llim, self.rlim = {}, {}


        self.adaptive =[]

        # sampling distribution
        self.pdf = {}
        self.cdf_inv = {}
        # prior
        self.prior_pdf = {}


        # histogram setup
        self.xpy = numpy
        self.identity_convert = lambda x: x  # if needed, convert to numpy format  (e.g, cupy.asnumpy)
        self.identity_convert_togpu = lambda x: x  # if needed, convert to numpy format  (e.g, cupy.asnumpy)

        # extra args, created during setup
        self.extra_args = {}

    def add_parameter(self, params, pdf,   left_limit=None, right_limit=None, prior_pdf=None, adaptive_sampling=False,**kwargs):
        """
        Add one (or more) parameters to sample dimensions. params is either a string describing the parameter, or a tuple of strings. The tuple will indicate to the sampler that these parameters must be sampled together. left_limit and right_limit are on the infinite interval by default, but can and probably should be specified. If several params are given, left_limit, and right_limit must be a set of tuples with corresponding length. Sampling PDF is required, and if not provided, the cdf inverse function will be determined numerically from the sampling PDF.
        """
        self.params.add(params) # does NOT preserve order in which parameters are provided
        self.params_ordered.append(params)
        # update dictionary limits
        if isinstance(params, tuple):
            assert all([lim[0] < lim[1] for lim in zip(left_limit, right_limit)])
            if left_limit is None:
                self.llim[params] = list(float("-inf"))*len(params)
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = list(float("+inf"))*len(params)
            else:
                self.rlim[params] = right_limit
        else:
            assert left_limit < right_limit
            if left_limit is None:
                self.llim[params] = float("-inf")
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = float("+inf")
            else:
                self.rlim[params] = right_limit
        self.pdf[params] = pdf
        self.prior_pdf[params] = prior_pdf

        if adaptive_sampling:
            print("   Adapting ", params)
            self.adaptive.append(params)


    def setup(self,  **kwargs):
        self.extra_args =kwargs  # may need to pass/use during the 'update' step

    def update_sampling_prior(self,ln_weights, n_history,tempering_exp=1,log_scale_weights=True,floor_integrated_probability=0,external_rvs=None,**kwargs):
      """
      update_sampling_prior

      Default setting is for lnL. The choices adopted are fairly arbitrary. For simplicity, we are using 'log scale weights' (derived from lnL, not L)
      which lets us sample well away from the posterior at the expense of not being as strongly sampling near the peak.

      NOTE: Currently deployed for mcsamplerPortfolio, NOT yet part of core code here
      """
      raise Exception(" - mcsampler_generic : not implemented: update_sampling_prior ")

    def draw_simplified(self,n_samples, *args, **kwargs):
        """
        draw n_samples

        Should not be implemented in generic class
        """
        #return joint_p_s, joint_p_prior, rv
        raise Exception(" - mcsampler_generic : not implemented: draw")

    def integrate(self, lnF, *args, xpy=xpy_default,**kwargs):
        use_lnL = kwargs['use_lnL'] if 'use_lnL' in kwargs else False
        raise Exception(" - mcsampler_generic : not implemented: integrate")
        

    def integrate_log(self, lnF, *args, xpy=xpy_default,**kwargs):
        raise Exception(" - mcsampler_generic : not implemented: integrate_log")
        xpy_here = self.xpy
        



#
# GOAL
#    Library to facilitate making various plots of equation-of-state-related quantities.


# from matplotlib import pyplot as plt
# import EOSManager
# import EOSPlotUtilities
# my_eos = EOSManager.EOSLALSimulation('SLy')
# EOSPlotUtilities.render_eos(my_eos.eos,'rest_mass_density', 'pressure')

import numpy as np
import RIFT.physics.EOSManager as EOSManager
import lalsimulation as lalsim
import lal

from matplotlib import pyplot as plt
try:
    from tqdm import tqdm
except:
    tqdm = lambda x:x

from scipy.interpolate import UnivariateSpline, PchipInterpolator


def render_eos(eos, xvar='energy_density', yvar='pressure',units='cgs',npts=100,label=None,logscale=True,verbose=False,**kwargs):

    min_pseudo_enthalpy = 0.005
    max_pseudo_enthalpy = lalsim.SimNeutronStarEOSMaxPseudoEnthalpy(eos)
    hvals = max_pseudo_enthalpy* 10**np.linspace( np.log10(min_pseudo_enthalpy/max_pseudo_enthalpy),  -1e-4,num=npts)
    if verbose:
        print(hvals,min_pseudo_enthalpy, max_pseudo_enthalpy)

    qry = EOSManager.QueryLS_EOS(eos)

    xvals = qry.extract_param(xvar,hvals)
    yvals = qry.extract_param(yvar,hvals)
    if verbose:
        print(np.c_[xvals,yvals])

    if logscale:
        plt.loglog(xvals, yvals,label=label,**kwargs)
    else:
        plt.plot(xvals, yvals,label=label,**kwargs)
    return None


def eval_eos_list_vs(eos_list, xvar='energy_density', xgrid=None,yvar='pressure', units='cgs',use_monotonic=True):
    if xgrid is None:
        raise Exception(" EOSPlotUtilities: none passed for grid")
    n_eos = len(eos_list)
    npts = len(xgrid)
#    print(npts,n_eos)
    # LARGE ALLOCATION potentially, so watch out -- usually I just need quantiles
    outvals  = np.zeros( (npts,n_eos))
    # loop and compute -- ideally parallelize! Silly to do serialy
    for indx in tqdm(np.arange(n_eos)):
        eos = eos_list[indx]
        # Pull out on grid
        min_pseudo_enthalpy = 0.005
        max_pseudo_enthalpy = lalsim.SimNeutronStarEOSMaxPseudoEnthalpy(eos)
        hvals = max_pseudo_enthalpy* 10**np.linspace( np.log10(min_pseudo_enthalpy/max_pseudo_enthalpy),  -1e-4,num=npts)
        qry = EOSManager.QueryLS_EOS(eos)   
        xvals = qry.extract_param(xvar,hvals)
        yvals = qry.extract_param(yvar,hvals)   
        # interpolate to target grid.   Usually interpolate log x to log y.  Assume INCREASING sample array. LINEAR interpolation
        #log_ygrid = np.interp(np.log(xgrid), np.log(xvals), np.log(yvals))
        if use_monotonic:
            intp_func = PchipInterpolator(np.log(xvals),np.log(yvals))
        else:
            intp_func = UnivariateSpline(np.log(xvals),np.log(yvals))
        ygrid = np.exp(intp_func(np.log(xgrid)))
        outvals[:,indx] = ygrid


    return outvals


def render_eos_list_quantiles_vs(eos_list, quantile_bounds=None, xvar='energy_density', xgrid=None,yvar='pressure', units='cgs',use_monotonic=True,use_log=True,return_outvals=False,input_outvals=None,show_traces=False,plot_kwargs={}, fill_kwargs={}):
    outvals_here=None
    if input_outvals is None:
        outvals_here  = eval_eos_list_vs(eos_list, xvar=xvar , xgrid=xgrid, yvar=yvar, units=units, use_monotonic=use_monotonic)
#        print(outvals_here[:,-1])
    else:
        outvals_here = input_outvals

    if outvals_here is None:
        raise Exception(" failure generating eval list, should never happen this way")

    xgrid_here = np.array(xgrid)
    upper_vals = np.percentile(outvals_here,quantile_bounds[0]*100,1)
    lower_vals = np.percentile(outvals_here,quantile_bounds[1]*100,1)
    if use_log:
        xgrid_here = np.log10(xgrid_here)
        upper_vals = np.log10(upper_vals)
        lower_vals = np.log10(lower_vals)

    if show_traces:
#        print(outvals_here.shape, xgrid_here.shape)
        for indx in np.arange(len(outvals_here)):
            if use_log:
                plt.plot(xgrid_here,np.log10(outvals_here[:,indx]),color='k')



    plt.plot(xgrid_here, upper_vals, **plot_kwargs)
    plt.plot(xgrid_here, lower_vals, **plot_kwargs)
    plt.fill_between(xgrid_here, lower_vals,upper_vals,**fill_kwargs)
    if return_outvals:
        return outvals
    return None


def render_reprimand_tovsequence_list_quantiles_vs(sequence_list, quantile_bounds=None, xvar='radius', ygrid=None,yvar='mass', range_mass = '[0.5,2.0]', steps = 1000, return_outvals=False,input_outvals=None, percentile_method = 'use nan percentile', plot_kwargs={}, fill_kwargs={}):
    import pyreprimand as pyr
    range_mass = eval(range_mass)
    mg = np.linspace(range_mass[0], range_mass[1], steps)
    rc = np.zeros( (len(mg),len(sequence_list)))
    
    for i in range(len(sequence_list)): rc[:,i] = sequence_list[i].circ_radius_from_grav_mass(mg) * sequence_list[i].units_to_SI.length/1e3
    
    # Remove EoSs which have unreasonably large radii from this analysis
    sequence_list = np.delete(sequence_list, np.where(rc> 50)[1])
    rc = np.delete(rc, np.where(rc > 50)[1], 1)
    
    if percentile_method == 'use nan percentile':
        ygrid_here = np.array(mg)
        lower_vals = np.nanpercentile(rc,quantile_bounds[0]*100,1)
        upper_vals = np.nanpercentile(rc,quantile_bounds[1]*100,1)
    elif percentile_method == 'regular percentile':
        '''Assumes NAN ==an infinitely large number. So all radius percentiles are NAN if even one Radius value at that Mass is a NAN. nanpercentile is better. https://www.geeksforgeeks.org/numpy-nanpercentile-in-python/#
        '''
        ygrid_here = np.array(mg)
        lower_vals = np.percentile(rc,quantile_bounds[0]*100,1)
        upper_vals = np.percentile(rc,quantile_bounds[1]*100,1)
    elif percentile_method == 'truncate curves':
        raise Exception("Not implemented yet.")
        for i in reversed(range(len(sequence_list))):
            if np.isnan(rc[-1,i]) :
                rc = np.delete(rc, i, 1)
        lower_vals = np.percentile(rc,quantile_bounds[0]*100,1)
        upper_vals = np.percentile(rc,quantile_bounds[1]*100,1)
        ygrid_here = np.array(mg)
    
    plt.plot(lower_vals, ygrid_here, **plot_kwargs)
    plot_kwargs['label'] = ''
    plt.plot(upper_vals, ygrid_here, **plot_kwargs)
    plt.fill_betweenx(ygrid_here, lower_vals,upper_vals,**fill_kwargs)
    if return_outvals:
        return mg, rc
    return None





#! /usr/bin/env python
# test_response_functions_Q.py
#    - Plot Qlm(t) [specifically, Q22] on response to a 'delta' function h(t)
#      By default, the 'delta' function is a gaussian
#    - Uses full test_like_and_samp.py infrastructure, so can read in data as well.
#
# USER MUST MANUALLY CHANGE
#   To turn on and off the discontinuity and windowing
# EXAMPLE
#  -  Plot response functions for the 'discontinuous' function plus signal, using a smooth iLIGO scalePSD
#         test_response_functions_Q.py --psd-file mypsd.xml.gz  
#         test_response_functions_Q.py --psd-file mypsd.xml.gz   --seglen 64  # repeat with more data
#
#   - Change the template durations, to move the signal around in time
#         test_response_functions_Q.py --psd-file mypsd.xml.gz   --seglen 64  # repeat with more data
#
#  - Repeat with an aLIGO-scale noisy (numerical) PSD
#         test_response_functions_Q.py --psd-file mypsd-noisy.xml.gz  --fmin-template 40 --signal-fmin 22 --show-psd --signal-dist 500     # use a noisy, numerical PSD

#  - Repeat for truncated inverse psd (e.g., noisy V1)
#    python test_response_functions_Q.py  --psd-file mypsd-noisy.xml.gz  --seglen 64 --fmin-template 30 --signal-fmin 22 --show-psd --psd-truncate-inverse --psd-truncate-inverse-time 8 --signal-dist 500
#    python test_response_functions_Q.py  --psd-file mypsd-noisy.xml.gz  --seglen 64  --show-psd --signal-fmin 22 --psd-truncate-inverse --psd-truncate-inverse-time 16 --signal-dist 500

#
#
#  - Reading in real data. [Typically, inj xml or coinc needed. Make sure seglen chosen *and* default instrument channel provided!]
#    Doing so, make *absolutely certain* the correct fref, Lmax, amporder are used!
#          python ../../Code/test_response_functions_Q.py --cache zero_noise.cache   --psd-file HLV-ILIGO_PSD.xml.gz --inj mdc.xml.gz  --fref 0 --Lmax 3 --amporder -1
#          python ../../Code/test_response_functions_Q.py --cache zero_noise.cache  --psd-file HLV-ILIGO_PSD.xml.gz --psd-truncate-inverse --psd-truncate-inverse-time 8 --inj mdc.xml.gz --fref 0 --Lmax 3 --amporder -1
#          python ../../Code/test_response_functions_Q.py --cache iligo_noise.cache  --psd-file HLV-ILIGO_PSD.xml.gz --psd-truncate-inverse --psd-truncate-inverse-time 8 --inj mdc.xml.gz --fref 0 --Lmax 3 --amporder -1
#          python ../../Code/test_response_functions_Q.py --cache iligo_noise.cache  --psd-file HLV-ILIGO_PSD.xml.gz --psd-truncate-inverse --psd-truncate-inverse-time 8 --inj mdc.xml.gz  --seglen 31 --padding 2  --fref 0 --Lmax 3 --amporder -1  #  basically HUGELY zero-pad data, to push ringing/response off. Not perfect
#         python ../../Code/test_response_functions_Q.py --cache iligo_noise.cache  --psd-file HLV-ILIGO_PSD.xml.gz  --channel V1=FAKE-STRAIN  --inj mdc.xml.gz  --seglen 31 --fref 0 --Lmax 3 --amporder -1 --show-likelihood-versus-time; open test-Q-* FLT-*
#        python ../../Code/test_response_functions_Q.py --cache iligo_noise.cache  --psd-file HLV-ILIGO_PSD.xml.gz   --channel V1=FAKE-STRAIN --channel H1=FAKE-STRAIN --channel L1=FAKE-STRAIN  --inj mdc.xml.gz  --seglen 31 --fref 0 --Lmax 3 --amporder -1 --psd-truncate-inverse --show-likelihood-versus-time --show-psd; open test-Q-* FLT-*

#        python ../../Code/test_response_functions_Q.py --cache s5_noise.cache  --psd-file HLV-ILIGO_PSD.xml.gz   --channel V1=FAKE-STRAIN --channel H1=FAKE-STRAIN --channel L1=FAKE-STRAIN  --inj mdc.xml.gz  --seglen 31 --fref 0 --Lmax 3 --amporder -1 --psd-truncate-inverse --show-likelihood-versus-time --show-psd; open test-Q-* FLT-*

from __future__ import print_function

try:
    import matplotlib
    #matplotlib.use("Agg")
    matplotlib.use("GDK")
    from matplotlib import pylab as plt
except:
    print("- no matplotlib -")


import lalsimutils
import numpy as np
import lal
import lalsimulation as lalsim

import factored_likelihood
import factored_likelihood_test

import sys
import functools
import pickle
import numpy

try:
    import healpy
    from lalinference.bayestar import fits as bfits
    from lalinference.bayestar import plot as bplot
except:
    print(" -no skymaps - ")


from glue.lal import Cache
from glue.ligolw import utils, lsctables, table, ligolw, git_version
from glue.ligolw.utils import process

import xmlutils
import ourparams
import ourio
opts,  rosDebugMessagesDictionary = ourparams.ParseStandardArguments()
factored_likelihood.rosDebugMessagesDictionary = rosDebugMessagesDictionary
lalsimutils.rosDebugMessagesDictionary            = rosDebugMessagesDictionary
print(opts)
print(rosDebugMessagesDictionary)

print(" ======= TESTING RESPONSE FUNCTION AND DISCONTINITY ====")
print("   Edit code variables 'bUseWindow' and 'bAddDiscontinity' to insert a plausible discontinuity ")
bUseWindow = True
bAddDiscontinuity = False
window_beta = 0.01
detDefault=  'V1'

def fakeGaussianDataTime(amp,sigma,deltaT, npts, window='Tukey', window_beta=0.01):
    # Create h(t)
    htGauss =lal.CreateCOMPLEX16TimeSeries("h", lal.LIGOTimeGPS(0.), 0., deltaT, lal.lalDimensionlessUnit, npts)
    htGauss.data.data[:] = 0.j  # initialize the data

    # Populate first few samples with a half-gaussian , +
    tvals = np.arange(npts)*deltaT
    htGauss.data.data += amp*np.exp(-tvals*tvals/(sigma*sigma)/2.)

    # Populate the last few samples with a half-gaussian, -
    tvalsMinus = np.arange(npts)[::-1]*deltaT
    htGauss.data.data += -amp*np.exp(-tvalsMinus*tvalsMinus/(sigma*sigma)/2.)

    if window:
        windowArray  = lal.CreateNamedREAL8Window(window, window_beta, len(htGauss.data.data))  # Window (start)
        htGauss.data.data *= windowArray.data.data

    return htGauss 
    
def DataFourier(ht):   # Complex fft wrapper (COMPLEX16Time ->COMPLEX16Freq. No error checking or padding!
    TDlen = ht.data.length
    fwdplan=lal.CreateForwardCOMPLEX16FFTPlan(TDlen,0)
    hf = lal.CreateCOMPLEX16FrequencySeries("Template h(f)", 
            ht.epoch, ht.f0, 1./ht.deltaT/TDlen, lal.lalHertzUnit, 
            TDlen)
    lal.COMPLEX16TimeFreqFFT(hf, ht, fwdplan)
    # assume memory freed by swig python
    return hf
def DataInverseFourier(hf):   # Complex fft wrapper (COMPLEX16Freq ->COMPLEX16Time. No error checking or padding!
    FDlen = hf.data.length
    dt = 1./hf.deltaF/FDlen
    revplan=lal.CreateReverseCOMPLEX16FFTPlan(FDlen,0)
    ht = lal.CreateCOMPLEX16TimeSeries("Template h(t)", 
            hf.epoch, hf.f0, dt, lal.lalDimensionlessUnit, 
            FDlen)
    lal.COMPLEX16FreqTimeFFT( ht, hf, revplan)  
    # assume memory freed by swig python
    return ht
## TEST and DEMO of truncate_inverse_psd
# fNyq = 2048
# df = 1/64
# psd = lalsimutils.get_psd_series_from_xmldoc("psd.xml.gz")
# psd = lalsimutils.resample_psd_series(psd, df, 0, fNyq)
# psd = lal.ResizeREAL8FrequencySeries(psd,0, fNyq/df+1)
# psdNew = lalsimutils.truncate_inverse_psd(psd,8,40,2000)   # impose hard limits at 40,2000 Hz
def truncate_inverse_psd(swig_psd,Tkeep,fmin,fmax):
    """
    Takes swig_psd, Tkeep,fmax.  Conputes inverse 1/psd^(1/2) (iFFT of sqrt(weights2side)) in the time domain, zeroing all except [-Tkeep/2, Tkeep/2]
    around t=0, inverse-fft's, and returns a 'revised' PSD, corresponding to a revised (finite-length) time duration filter.   
   *Requires* PSD  initialized up to its nyquist frequency, in swig format, with pow2+1  entries.

    *fmin* and *fmax* are supplied explicitly by the user, to *ensure* any hard cutoffs are explicitly included in the PSD model,
    using *exactly* the same code paths used elsewhere/previously
        [Even without lines, the low-frequency cutoff introduces long-duration 'ringing' into the whitening filter, which this program is intended to truncate]
        [*WARNING*: This means that in *subsequent* uses of this PSD, you need to set flow = 0 ! ]

    Implemented using 'IP' infrastructure  (IP.intgd = freq; IP.ovlp = time) to allocate memory and map the PSD arrays,
    to insure consistent length requirements and bin alignment (i.e., code path) as implementation used in inner product.
    (Ideally, should probably use IP to return the inverse FFT too)

    For comparison: https://www.lsc-group.phys.uwm.edu/daswg/projects/lal/nightly/docs/html/_find_chirp_s_p_data_8c_source.html#l00431
    Reference: Section 4.7 in Duncan's thesis http://arxiv.org/pdf/0705.1514v1.pdf 
    """

    # Create frequencies : df, fNyq
    df = swig_psd.deltaF
    fNyq = df * (len(swig_psd.data.data) -1)   # Implicit assumption about odd length of raw psd file
    T = 1/df  # = N deltaT
    # We cannot truncate if the timescale is too short. Return untruncated PSD (not safe; perhaps die?)
    if T<1.5*Tkeep:
        return swig_psd
    # Proceed: make safety copy of swig_psd, to prevent accidentally changing it by numpy magic (e.g., side effects in IP)
    swig_psd_copy = lal.CutREAL8FrequencySeries(swig_psd,0, len(swig_psd.data.data)) # swig passes pointers, don't change the original array by accident

    # Create, extract 'weights2side' [PSD in array form] and two pre-allocated buffers
#    print " Creating IP structure, truncating the PSD over the range ", [fmin, fmax]
    # Master branch arguments
    IP = lalsimutils.ComplexOverlap(fLow=fmin, fMax=fmax,fNyq=fNyq, deltaF=swig_psd_copy.deltaF,psd=swig_psd_copy.data.data, analyticPSD_Q=False)
    npts = len(IP.intgd.data.data)  # = IP.len2side
    wt = lal.CutCOMPLEX16FrequencySeries(IP.intgd,0,npts)  # instantiate 2-sided freq series, copied from IP.  [associated timeseries in 'IP.ovlp'-> xt]
    wt.data.data = np.sqrt(IP.weights2side) # populate with weights (=1/sqrt(PSD)].  This includes any truncation (flow, fmax, etc)

    # SCALE FACTOR
    #   - concern we did not keep enough numerical precision to preserve zeros
    wtScale = np.abs(np.max(wt.data.data[np.nonzero(wt.data.data)]))
    wt.data.data = wt.data.data/wtScale
    # Inverse FFT to create timeseries (ROS wrapper)
    xt = DataInverseFourier(wt)

    # Truncate the series, targeting a specific duration.  (Up to user to make sure these durations are sane)
    dt =xt.deltaT
    nSamplesToZero = int((T-Tkeep)/dt/2)  # The ifft array in *time* has t=0 at the first sample. Zero out the middle 2*nSamplesToZero points
    nSamples = xt.data.length
    xt.data.data[nSamplesToZero+1:nSamples-nSamplesToZero] = 0.   # Be careful re potential fencepost error here.
    # FFT the truncated weights
    wtNew = DataFourier(xt)


    # Allocate and populate a *one-sided* REAL8FrequencySeries, of correct length. [Remember, we assume fully allocated PSD at start]
    # Avoid direct numpy assignment, since mixing memory models produces segfaults
    swig_psd_out = lal.CutREAL8FrequencySeries(swig_psd,0, swig_psd.data.length)  # make *another* copy, to write into
    #swig_psd_out.data.data = np.zeros(len(swig_psd_out.data.data))
    tmp =  wtNew.data.data[0:npts/2+1]  # Assign data from first half, including f=0 bin. BE CAREFUL re potential fenceposts
    
    tmp = tmp[::-1]                               # Reverse so f=0 is first bin. Remember f=0 is one past midpoint in original array
    tmpSquare = np.abs(tmp)**2            #  create 1/S^2 array. Because the FFT is 2-sided, there is a small complex part created. Use np.abs() to force real.
    for i in np.arange(swig_psd_out.data.length - 1):
        swig_psd_out.data.data[i]=0
        if tmpSquare[i]:   # avoid underflow
            swig_psd_out.data.data[i] = 1./tmpSquare[i]  # assign
    # Set the zero and nyquist bin
    swig_psd_out.data.data[0] = np.inf
    swig_psd_out.data.data[-1] =0.
    # Apply scale factor at end
    swig_psd_out.data.data *= 1/np.power(wtScale,2)  # re-insert scale factor. Remember wt scales as longweights. This tries to keep numbers in the FFT's of order unity, to avoid precision loss

    # Return truncated PSD
    print("   Truncated PSD created ")
    return swig_psd_out


def fakeGaussianDataFrequency(amp, sigma,deltaT,npts,window='Tukey', window_beta=0.01):
    return DataFourier(fakeGaussianDataTime(amp,sigma,deltaT,npts,window=window,window_beta=window_beta))

if opts.verbose:
    try:
        from subprocess import call
        print(" ---- LAL Version ----")
        call(["lal-version"])
        print(" ---- GLUE Version ----")
        call(["ligolw_print",  "--version"])
        print(" ---- pylal Version ----")
        call(["pylal_version"])
    except:
        print("  ... trouble printing version numbers ")
    print(" Glue tag ", git_version.id)
    try:
        import scipy   # used in lalsimutils; also printed in pylal-version
        print(" Scipy: ", scipy.__version__)
    except:
        print(" trouble printing scipy version")

def mean_and_dev(arr, wt):
    av = np.average(arr, weights=wt)
    var = np.average(arr*arr, weights=wt)
    return [av, var - av*av]

def pcum_at(val,arr, wt):
    nm = np.sum(wt)
    return np.sum(wt[np.where(val < arr)]/nm)

__author__ = "R. O'Shaughnessy"


checkInputs = opts.plot_ShowLikelihoodVersusTime

rosUseDifferentWaveformLengths = False    
rosUseRandomTemplateStartingFrequency = False

rosUseTargetedDistance = True
rosUseStrongPriorOnParameters = False
rosShowSamplerInputDistributions = opts.plot_ShowSamplerInputs
rosShowRunningConvergencePlots = True
rosShowTerminalSampleHistograms = True
rosUseMultiprocessing= False
rosDebugCheckPriorIntegral = False
nMaxEvals = int(opts.nmax)
print(" Running at most ", nMaxEvals, " iterations")

fracThreshold = opts.points_threshold_match

theEpochFiducial = lal.LIGOTimeGPS(1064023405.000000000)   # 2013-09-24 early am 
tEventFiducial = 0                                                                 # relative to GPS reference


det_dict = {}
rhoExpected ={}

tWindowReference =   factored_likelihood_test.tWindowReference
tWindowExplore =     factored_likelihood.tWindowExplore

approxSignal = lalsim.GetApproximantFromString(opts.approx)
approxTemplate = approxSignal
ampO =opts.amporder # sets which modes to include in the template (and signal, if injected)
Lmax = opts.Lmax # sets which modes to include in the template.  Print warning if inconsistent.
fref = opts.fref
fminWavesTemplate = opts.fmin_Template  # too long can be a memory and time hog, particularly at 16 kHz
fminWavesSignal = opts.signal_fmin  # too long can be a memory and time hog, particularly at 16 kHz
fminSNR =opts.fmin_SNR
fmaxSNR =opts.fmax_SNR
fSample = opts.srate

if ampO ==-1 and Lmax < 5:
    print(" +++ WARNING ++++ ")
    print("  : Lmax is ", Lmax, " which may be insufficient to resolve all higher harmonics in the signal! ")

if (ampO+2)> Lmax:
    print(" +++ WARNING ++++ ")
    print("  : Lmax is ", Lmax, " which may be insufficient to resolve all higher harmonics in the signal! ")

if opts.channel_name is not None and opts.cache_file is None:
    print("Cache file required when requesting channel data.", file=sys.stderr)
    exit(-1)
elif opts.channel_name is not None:
    det_dict = dict(map(lambda cname: cname.split("="), opts.channel_name))




#
# Produce data with a coherent signal in H1, L1, V1
#
data_dict = {}
Psig = None

# Read in *coincidence* XML (overridden by injection, if present)
if opts.coinc:
    print("Loading coinc XML:", opts.coinc)
    xmldoc = utils.load_filename(opts.coinc)
    coinc_table = table.get_table(xmldoc, lsctables.CoincInspiralTable.tableName)
    assert len(coinc_table) == 1
    coinc_row = coinc_table[0]
    event_time = float(coinc_row.get_end())  # 
    event_time_gps = lal.GPSTimeNow()    # Pack as GPSTime *explicitly*, so all time operations are type-consistent
    event_time_gps.gpsSeconds = int(event_time)
    event_time_gps.gpsNanoSeconds = int(1e9*(event_time -event_time_gps.gpsSeconds))
    theEpochFiducial = event_time_gps       # really should avoid duplicate names
    print("Coinc XML loaded, event time: %s" % str(coinc_row.get_end()))
    # Populate the SNR sequence and mass sequence
    sngl_inspiral_table = table.get_table(xmldoc, lsctables.SnglInspiralTable.tableName)
    m1, m2 = None, None
    for sngl_row in sngl_inspiral_table:
        # NOTE: gstlal is exact match, but other pipelines may not be
        assert m1 is None or (sngl_row.mass1 == m1 and sngl_row.mass2 == m2)
        m1, m2 = sngl_row.mass1, sngl_row.mass2
        rhoExpected[str(sngl_row.ifo)] = sngl_row.snr  # record for comparisons later
        if rosDebugMessagesDictionary["DebugMessages"]:
            det = str(sngl_row.ifo)
            print(det, rhoExpected[det])
    m1 = m1*lal.LAL_MSUN_SI
    m2 = m2*lal.LAL_MSUN_SI
    rho2Net = 0
    for det in rhoExpected:
        rho2Net += rhoExpected[det]**2
    if rosDebugMessagesDictionary["DebugMessages"]:
        print(" Network :", np.sqrt(rho2Net))
    # Create a 'best recovered signal'
    Psig = lalsimutils.ChooseWaveformParams(
        m1=m1,m2=m2,approx=approxSignal,
        fmin = fminWavesSignal, 
        dist=factored_likelihood.distMpcRef*1e6*lal.LAL_PC_SI,    # default distance
        fref=fref, 
        tref = event_time_gps,
        ampO=ampO      
        )  # FIXME: Parameter mapping from trigger space to search space
    if rosDebugMessagesDictionary["DebugMessages"]:
        print(" === Coinc table : estimated signal [overridden if injection] ===")
        Psig.print_params()

# Read in *injection* XML
if opts.inj:
    print("Loading injection XML:", opts.inj)
    Psig = lalsimutils.xml_to_ChooseWaveformParams_array(str(opts.inj))[opts.event_id]  # Load in the physical parameters of the injection.  
    m1 = Psig.m1
    m2 = Psig.m2
    timeWaveform = float(-lalsimutils.hoft(Psig).epoch)
    Psig.deltaF = 1./lalsimutils.nextPow2(opts.seglen)       # Frequency binning needs to account for target segment length
    theEpochFiducial = Psig.tref  # Reset
    tEventFiducial = 0               # Reset
    print(" ++ Targeting event at time ++ ", lalsimutils.stringGPSNice(Psig.tref))
    print(" +++ WARNING: ADOPTING STRONG PRIORS +++ ")
    rosUseStrongPriorOnParameters= True
    Psig.print_params()
    print("---- End injeciton parameters ----")

# Use forced parameters, if provided
if opts.template_mass1:
    Psig.m1 = opts.template_mass1*lal.LAL_MSUN_SI
if opts.template_mass2:
    Psig.m2 = opts.template_mass2*lal.LAL_MSUN_SI

# Reset origin of time, if required. (This forces different parts of data to be read- important! )
if opts.force_gps_time:
    print(" +++ USER HAS OVERRIDDEN FIDUCIAL EPOCH +++ ")
    print("  The zero of time (and the region to be windowed) will be changed; you had better know what you are doing.  ")
    print("    original " ,lalsimutils.stringGPSNice(theEpochFiducial))
    print("    new      ", opts.force_gps_time)
    theEpochFiducial = lal.GPSTimeNow()
    theEpochFiducial.gpsSeconds = int(opts.force_gps_time)
    theEpochFiducial.gpsNanoSeconds =  int(1e9*(opts.force_gps_time - int(opts.force_gps_time)))

# Create artificial "signal".  Needed to minimize duplicate code when I
#  -  consistently test waveform duration
#  - copy template parameters 
#  - [test code] : generate plots vs time [test code], expected SNR, etc
#       WARNING: Test code plots will not look perfect, because we don't know the true sky location (or phase, polarization, ...)
if  not Psig and opts.channel_name:  # If data loaded but no signal generated
    if (not opts.template_mass1) or (not opts.template_mass2) or (not opts.force_gps_time):
        print(" CANCEL: Specifying parameters via m1, m2, and time on the command line ")
        sys.exit(0)
    Psig = lalsimutils.ChooseWaveformParams(approx=approxSignal,
        fmin = fminWavesSignal, 
        dist=factored_likelihood.distMpcRef*1e6*lal.LAL_PC_SI,    # default distance
        fref=fref)
    Psig.m1 = lal.LAL_MSUN_SI*opts.template_mass1
    Psig.m2 = lal.LAL_MSUN_SI*opts.template_mass2
    Psig.tref = lal.LIGOTimeGPS(0.000000000)  # Initialize as GPSTime object
    Psig.tref += opts.force_gps_time # Pass value of float into it


# TEST THE SEGMENT LENGTH TARGET
if Psig:
    timeSegmentLength  = float(-lalsimutils.hoft(Psig).epoch)
    if rosDebugMessagesDictionary["DebugMessages"]:
        print(" Template duration : ", timeSegmentLength)
    if timeSegmentLength > opts.seglen:
        print(" +++ CATASTROPHE : You are requesting less data than your template target needs!  +++")
        print("    Requested data size: ", opts.seglen)
        print("    Template duration  : ", timeSegmentLength)
        sys.exit(0)


# TRY TO READ IN DATA: if data specified, use it and construct the detector list from it. Otherwise...
if opts.channel_name and    (opts.opt_ReadWholeFrameFilesInCache):
    for inst, chan in map(lambda c: c.split("="), opts.channel_name):
        print("Reading channel %s from cache %s" % (inst+":"+chan, opts.cache_file))
        data_dict[inst] = lalsimutils.frame_data_to_non_herm_hoff(opts.cache_file, inst+":"+chan,window_shape=window_beta)
        fSample = len(data_dict[inst].data.data)*data_dict[inst].deltaF
        df = data_dict[inst].deltaF
        if Psig:
            Psig.deltaF = df
        print("Frequency binning: %f, length %d" % (data_dict[inst].deltaF, len(data_dict[inst].data.data)))
        print("Sampling rate ", fSample)
    detDefault = (data_dict.keys())[0]
if opts.channel_name and not (opts.opt_ReadWholeFrameFilesInCache):
    if Psig:
        event_time = Psig.tref
    else:
        event_time = theEpochFiducial  # For now...get from XML if that is the option
    start_pad, end_pad = opts.seglen-opts.padding, opts.padding 
    for inst, chan in map(lambda c: c.split("="), opts.channel_name):
        print("Reading channel %s from cache %s" % (inst+":"+chan, opts.cache_file))
        # FIXME: Assumes a frame file exists covering EXACTLY the needed interval!
        taper = lalsim.LAL_SIM_INSPIRAL_TAPER_STARTEND
        data_dict[inst] = lalsimutils.frame_data_to_non_herm_hoff(opts.cache_file, inst+":"+chan, start=int(event_time)-start_pad, stop=int(event_time)+end_pad,window_shape=window_beta)
        fSample = len(data_dict[inst].data.data)*data_dict[inst].deltaF
        df = data_dict[inst].deltaF
        if Psig:
            Psig.deltaF =df
        print("Frequency binning: %f, length %d" % (data_dict[inst].deltaF, len(data_dict[inst].data.data)))
        print("Sampling rate ", fSample)

    detDefault = (data_dict.keys())[0]
#        print " Sampling rate of data ", fSample

# CREATE A DEFAULT "signal", if none made to this point.  
analytic_signal = False
if len(data_dict) is 0:
    analytic_signal = True

    if not(Psig):
        m1 = 4*lal.LAL_MSUN_SI
        m2 = 3*lal.LAL_MSUN_SI

        Psig = lalsimutils.ChooseWaveformParams(
            m1 = m1,m2 =m2,
            fmin = fminWavesSignal, 
            fref=fref, 
            tref = theEpochFiducial + tEventFiducial,
            approx=approxSignal,
            ampO=ampO,
            radec=True, theta=1.2, phi=2.4,
            detector='H1', 
            dist=opts.signal_distMpc*1.e6*lal.LAL_PC_SI,    # move back to aLIGO distances
            deltaT=1./fSample
                                )
        timeSegmentLength  = -float(lalsimutils.hoft(Psig).epoch)
        if timeSegmentLength > opts.seglen:
            print(" +++ CATASTROPHE : You are requesting less data than your template target needs!  +++")
            print("    Requested data size: ", opts.seglen)
            print("    Template duration  : ", timeSegmentLength)
            sys.exit(0)

    if opts.signal_incl:
        Psig.incl = opts.signal_incl
    if opts.signal_psi:
        Psig.psi = opts.signal_psi
    if opts.signal_phiref:
        Psig.phi = opts.signal_phiref
    if opts.signal_tref:
        if np.abs(opts.signal_tref) < 1000:
            Psig.tref = theEpochFiducial+opts.signal_tref  
        else:
            Psig.tref = opts.signal_tref
            theEpochFiducial=opts.signal_tref

    df = lalsimutils.findDeltaF(Psig)
    if 1/df < opts.seglen:   # Allows user to change seglen of data for *analytic* models, on the command line. Particularly useful re testing PSD truncation
        df = 1./lalsimutils.nextPow2(opts.seglen)
    Psig.print_params()
    Psig.deltaF = df

    # Create gaussian fake data
    fNyqTarget = 2048.
    npts = 2*int(fNyqTarget/df)
    amp = 1e-15
    window_beta = window_beta
    if not bAddDiscontinuity:
        amp *=0
    if bUseWindow:
        # create signal first, to set epoch
        data_dict[detDefault] = lalsimutils.non_herm_hoff(Psig)
        sig = fakeGaussianDataFrequency(amp,2, 1/(2*fNyqTarget),npts,window='Tukey',window_beta=window_beta)
    else:
        data_dict[detDefault] = lalsimutils.non_herm_hoff(Psig)
        sig = fakeGaussianDataFrequency(amp,2, 1/(2*fNyqTarget),npts)

    data_dict[detDefault].data.data += sig.data.data


    # Report on fake data
    print(" ============")
    print(" Fake data report : (n,df,fNyq) ", data_dict[detDefault].data.length, data_dict[detDefault].deltaF, data_dict[detDefault].deltaF*data_dict[detDefault].data.length/2)



# PSD reading
psd_dict = {}
if not(opts.psd_file):
    analyticPSD_Q = True # For simplicity, using an analytic PSD
    for det in data_dict.keys():
        psd_dict[det] = lal.LIGOIPsd # lalsim.SimNoisePSDaLIGOZeroDetHighPower   #lal.LIGOIPsd    # Preserves key equality in data_dict , psd_dict

else:
    analyticPSD_Q = False # For simplicity, using an analytic PSD
    detectors = data_dict.keys()
    df = data_dict[detectors[0]].deltaF
    fNyq = (len(data_dict[detectors[0]].data.data)/2)*df
    print(" == Loading numerical PSD ==")
    for det in detectors:
        print("Reading PSD for instrument %s from %s" % (det, opts.psd_file))

        # "Standard" PSD parsing code used on master.
        psd_dict[det] = lalsimutils.get_psd_series_from_xmldoc(opts.psd_file, det)  # pylal type!
        tmp = psd_dict[det].data
        print("Sanity check reporting : pre-extension, min is ", np.min(tmp), " and maximum is ", np.max(tmp))
        deltaF = data_dict[det].deltaF
        fmin = psd_dict[det].f0
        fmax = fmin + psd_dict[det].deltaF*len(psd_dict[det].data)-deltaF
        print("PSD deltaF before interpolation %f" % psd_dict[det].deltaF)
        psd_dict[det] = lalsimutils.resample_psd_series(psd_dict[det], deltaF)
        print("PSD deltaF after interpolation %f" % psd_dict[det].deltaF)
        print("Post-extension the new PSD has 1/df = ", 1./psd_dict[det].deltaF, " (data 1/df = ", 1./deltaF, ") and length ", len(psd_dict[det].data.data))
        tmp = psd_dict[det].data.data
        nBad = np.argmin(tmp[np.nonzero(tmp)])
        fBad = nBad*deltaF
        print("Post-extension sanity check reporting  : min is ", np.min(tmp[np.nonzero(tmp)]), "(at n=", np.argmin(tmp[np.nonzero(tmp)])," or f=", fBad, ")  and maximum is ", np.max(psd_dict[det].data.data))
        print()

        # Inverse PSD truncation (optional)
        if opts.psd_TruncateInverse:
            Tleft = opts.psd_TruncateInverseTime
            print("  Attempting to truncate inverse PSD, using target length ", Tleft)
            psd_dict[det] = truncate_inverse_psd(psd_dict[det],Tleft,opts.fmin_SNR,opts.fmax_SNR)
            print("  Confirm returned PSD is sane " , psd_dict[det].data.data[0], psd_dict[det].data.data[-1])
            
            print(" To use this PSD *precisely* will require changing fminSNR and fmaxSNR ")
            fminSNR=5
            print("   SNR integrand ", [fminSNR, fNyq])


        #ARCHIVAL: Original non-interpolated method. Seems to fail asymptomatically (memory management?) for some PSD files
        # DANGER: Assumes PSD file starts at 0 hz
        # psd_dict[det] = lalsimutils.pylal_psd_to_swig_psd(lalsimutils.get_psd_series_from_xmldoc(opts.psd_file, det))
        # psd_dict[det] = lalsimutils.regularize_swig_psd_series_near_nyquist(psd_dict[det], fNyq-opts.fmax_SNR) # zero out 80 hz window near nyquist
        # psd_dict[det] =  lalsimutils.enforce_swig_psd_fmin(psd_dict[det], fminSNR)           # enforce fmin at the psd level, HARD CUTOFF
        # tmp = psd_dict[det].data.data
        # print "Sanity check reporting : pre-extension, min is ", np.min(tmp[np.nonzero(tmp)]), " and maximum is ", np.max(psd_dict[det].data.data)
        # deltaF = data_dict[det].deltaF
        # # remember the PSD is one-sided, but h(f) is two-sided. The lengths are not equal.
        # psd_dict[det] = lalsimutils.extend_swig_psd_series_to_sampling_requirements(psd_dict[det], df, df*(len(data_dict[det].data.data)/2))
        # print "Post-extension the new PSD has 1/df = ", 1./psd_dict[det].deltaF, " (data 1/df = ", 1./deltaF, ") and length ", len(psd_dict[det].data.data)
        # tmp = psd_dict[det].data.data
        # nBad = np.argmin(tmp[np.nonzero(tmp)])
        # fBad = nBad*deltaF
        # print "Post-extension sanity check reporting  : min is ", np.min(tmp[np.nonzero(tmp)]), "(at n=", np.argmin(tmp[np.nonzero(tmp)])," or f=", fBad, ")  and maximum is ", np.max(psd_dict[det].data.data)
        # print

# This code is a DUPLICATE TEST, used to initialize the peak log likelihood.
# I use this threshold to identify points for further investigation.
detectors = data_dict.keys()
if False: #not (opts.coinc) :
    print(" == Data report [USELESS IF NOISY DATA] == ")
    rho2Net = 0
    print(" Amplitude report *from data*:")
    for det in detectors:
        if analyticPSD_Q:
            IP = lalsimutils.ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det])
        else:
            IP = lalsimutils.ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[det].data.data,analyticPSD_Q=False)
        rhoExpected[det] = rhoDet = IP.norm(data_dict[det])
        rho2Net += rhoDet*rhoDet
        print(det, " rho = ", rhoDet)
    print("Network : ", np.sqrt(rho2Net))

if opts.plot_ShowH: 
    print(" == Plotting *raw* detector data data (time domain) via MANUAL INVERSE FFT  == ")
    print("     [This plot includes any windowing and data padding]   ")
    for det in detectors:
        revplan = lal.CreateReverseCOMPLEX16FFTPlan(len(data_dict[det].data.data), 0)
        ht = lal.CreateCOMPLEX16TimeSeries("ht", 
                lal.LIGOTimeGPS(0.), 0., Psig.deltaT, lal.lalDimensionlessUnit,
                len(data_dict[det].data.data))
        lal.COMPLEX16FreqTimeFFT(ht, data_dict[det], revplan)

        tvals = float(ht.epoch - theEpochFiducial) + ht.deltaT*np.arange(len(ht.data.data))  # Not correct timing relative to zero - FIXME
        plt.figure(1)
        plt.plot(tvals, np.abs(ht.data.data),label=det)
    plt.legend()
    plt.show()

    print(" == Plotting detector data (time domain; requires regeneration, MANUAL TIMESHIFTS,  and seperate code path! Argh!) == ")
    P = Psig.copy()
    P.tref = Psig.tref
    for det in detectors:
        P.detector=det 
        hT = lalsimutils.complex_hoft(P)   # Note detector time propagation is NOT performed by this function. This is NOT a fair comparison!
        tvals = float(P.tref - theEpochFiducial) + hT.deltaT*np.arange(len(hT.data.data))  # Not correct timing relative to zero - FIXME
        plt.figure(1)
        plt.plot(tvals, np.abs(hT.data.data),label=det)

    tlen = hT.deltaT*len(np.nonzero(np.abs(hT.data.data)))
#    tRef = np.abs(float(hT.epoch))
#    plt.xlim( tRef-0.5,tRef+0.1)  # Not well centered, based on epoch to identify physical merger time
#    plt.savefig("test_like_and_samp-input-hoft.pdf")
    plt.show()

# Load skymap, if present
if opts.opt_UseSkymap:
    print(" ==Loading skymap==")
    print("   skymap file ", opts.opt_UseSkymap)
    smap, smap_meta = bfits.read_sky_map(opts.opt_UseSkymap)
    sides = healpy.npix2nside(len(smap))

    if opts.plot_ShowSamplerInputs:
        try:
            from lalinference.bayestar import plot
            plt.subplot(111, projection='astro mollweide')
            plot.healpix_heatmap(smap)
            plt.show()
        except:
            print(" No skymap for you ")


#     import bisect

#     # WARNING: smap defined here and 'globally' stored
#     # FOR THE RECORD:http://healpix.sourceforge.net/pdf/intro.pdf for angle conversions
#     cum_smap = np.cumsum(smap)
#     def bayestar_cdf_inv(x,y):                 # it will be passed two arguments, use one
#         indx = bisect.bisect(cum_smap,x) 
#         th,ph = healpy.pix2ang(sides, indx)
#         if rosDebugMessagesDictionary["DebugMessagesLong"]:
#             print " skymap used x->(th,ph) :", x,th,ph
#         return ph,th-np.pi/2.

# #    bayestar_cdf_inv_vector = np.vectorize(bayestar_cdf_inv)
#     def bayestar_cdf_inv_vector(x,y):   # Manually vectorize, so I can insert pdb breaks
# #        pdb.set_trace()
#         indxVec = map(lambda z: bisect.bisect(cum_smap,z), x) 
#         th, ph = healpy.pix2ang(sides,indxVec)
#         return np.array([ph, th-np.pi/2.])                            # convert to RA, DEC from equatorial. Return must be pair of arrays

#     def bayestar_pdf_radec(x,y):               # look up value at targted pixel 
#         ra,dec = x,y
# #        pdb.set_trace()
#         indx = bplot._healpix_lookup(smap,  ra, dec) # note bplot takes lon (=ra, really), lat (dec). Not sure why
#         return smap[indx]


# Struct to hold template parameters
P = lalsimutils.ChooseWaveformParams(fmin=fminWavesTemplate, radec=False, incl=0.0,phiref=0.0, theta=0.0, phi=0,psi=0.0,
         m1=m1,m2=m2,
         ampO=ampO,
         approx=approxTemplate,
         fref=fref,
         tref=theEpochFiducial,
         deltaT=1./fSample,
         deltaF=df)

if opts.template_mass1:
    P.m1 = lal.LAL_MSUN_SI*opts.template_mass1
if opts.template_mass2:
    P.m2 = lal.LAL_MSUN_SI*opts.template_mass2

timeWaveformTemplate = float(-lalsimutils.hoft(P).epoch)
timeWaveform             = float(-lalsimutils.hoft(Psig).epoch)
print(" Reminder: signal duration, template duration, PSD truncQ, psdTruncT ", timeWaveform, timeWaveformTemplate) #, opts.psd_TruncateInverse, opts.psd_TruncateInverseTime


#
# Perform the Precompute stage
#    - use a large tWindow : I want to retain the *entire* timeseries
#    - will it work if tWindowReference[1] is large
#
#tWindowReference[1] = Psig.deltaT*len(data_dict[detectors[0]].data.data)
#print tWindowReference[1]
print("  ---- > Problem : Master automatically clips the timeseries duration; use a *hardcoded* large window, for now, to retain as much as plausible ")
print("          Note: if the window is too large, you hit t_shift < 0 ")
det0 = detectors[0]
tWindowReference[1]= timeWaveformTemplate # data_dict[det0].data.epoch - 
rholms_intp, crossTerms, rholms = factored_likelihood.PrecomputeLikelihoodTerms(theEpochFiducial,tWindowReference[1], P, data_dict,psd_dict, Lmax, fmaxSNR, analyticPSD_Q)
print("Done with precompute")
print(detectors, rholms.keys())

# Plot h(t)
plt.figure(0)
for det in detectors:
    ht = DataInverseFourier(data_dict[det])  # this had better be windowed, if the input is!
    tvals = np.arange(ht.data.length)*ht.deltaT + float(ht.epoch - theEpochFiducial)
    plt.plot(tvals,np.real(ht.data.data),label='h(t):'+det)
plt.xlabel('t(s): 0 at ' +  lalsimutils.stringGPSNice(theEpochFiducial))
plt.ylabel('h (data)')
plt.legend()
plt.savefig("test-Q-response-ht.jpeg")

#
# Plot rho22(t)
#
plt.figure(1)
plt.clf()
for det in detectors:
    rho22 =rholms[det][(2,2)]
    tvals = np.arange(rho22.data.length)*rho22.deltaT + float(rho22.epoch -theEpochFiducial) # use time alignment
    plt.plot(tvals,np.abs(rho22.data.data),label='rho22:'+det)
plt.xlabel('t(s) [relative]:  0 at ' +  lalsimutils.stringGPSNice(theEpochFiducial))
plt.ylabel('Q22')
plt.legend()
plt.savefig("test-Q-response-rho22.jpeg")


plt.clf()
plt.figure(2)
rho22 =rholms[detDefault][(2,2)]
tvals = np.arange(rho22.data.length)*rho22.deltaT + float(rho22.epoch -theEpochFiducial) # use time alignment
plt.plot(tvals,np.log10(np.abs(rho22.data.data)),label='logrho22')
plt.xlabel('t(s)')
plt.ylabel('log Q22')
plt.legend()
plt.savefig("test-Q-response-log-rho22.jpeg")

#
# Plot lnLData(t). BE CAREFUL: it is rolled, make sure time identifications correct
#
plt.clf()
tStartOffsetDiscrete = tvals[0]
print(" Plotting lnLData from ", float(theEpochFiducial+tStartOffsetDiscrete), " onward ")
for det in detectors:
    nBinsDiscrete = len(data_dict[det].data.data)
    lnLDataDiscrete = factored_likelihood.DiscreteSingleDetectorLogLikelihoodData(theEpochFiducial,rholms, theEpochFiducial+tStartOffsetDiscrete, nBinsDiscrete, Psig.phi, Psig.theta, Psig.incl, Psig.phiref,Psig.psi, Psig.dist, Lmax, det)
    plt.plot(tvals[:len(lnLDataDiscrete)],lnLDataDiscrete,label='lnLData:'+det)
plt.xlabel('t [fixme]  0 at ' +  lalsimutils.stringGPSNice(theEpochFiducial))
plt.ylabel('lnLData')
plt.savefig("test-Q-response-lnLData.jpeg")
plt.xlim(-0.5, 0.5)
plt.savefig("test-Q-response-lnLData-Zoom.jpeg")

#
# Plot lnL(t) around the event (interpolated, remember)
#
nptsMore = int(fSample*(tWindowExplore[1]-tWindowExplore[0]))
print(nptsMore)
tvals = np.linspace(tWindowExplore[0],tWindowExplore[1], nptsMore)
lnL = np.zeros(len(tvals))
for indx in np.arange(len(tvals)):
    # Make sure to populate the 'passed' parameters consistent with the injected value (sky location, etc)
    P.theta = Psig.theta
    P.phi   = Psig.phi
    P.dist   = Psig.dist
    P.incl   = Psig.incl
    P.psi    = Psig.psi  
    P.phiref = Psig.phiref
    P.tref =  theEpochFiducial+tvals[indx]
    lnL[indx] =  factored_likelihood.FactoredLogLikelihood(P, rholms_intp, crossTerms, Lmax)
plt.clf()
plt.plot(tvals,lnL,label='lnL')
plt.xlabel('t - tEvent (s): relative to '+lalsimutils.stringGPSNice(theEpochFiducial))
plt.ylabel('lnL')
plt.savefig("test-Q-response-lnL.jpg")

#
# Plot PSD
#
if opts.plot_ShowPSD:
    nPSD=0
    plt.clf()
    for det in psd_dict.keys():
        nPSD+=0  # Useful offset condition, if the PSDs are too closely spaced
        if analyticPSD_Q:
            fvals =  np.arange(opts.fmin_SNR,2000,1.)              # the upper limit is kind of random
            Sh   =  map(psd_dict[det],fvals)
        else:
            fvals = np.arange(len(psd_dict[det].data.data))*deltaF  # stored as one-sided PSD
            Sh  = psd_dict[det].data.data
            nSkip = int(len(fvals)/(4096))
            fvals = fvals[::nSkip]   # downsample! PSDs are often too huge to plot!
            Sh = Sh[::nSkip]        # downsample!
        plt.figure(0)
        plt.plot(np.log10(fvals),np.log10(Sh)+nPSD,label="Sh:"+det)
        plt.xlabel('f (Hz)')
        plt.ylabel('Sh $Hz^{-1}$')
        plt.title('PSDs used')
    # Comparison plot: iLIGO
    Sh = map(lal.LIGOIPsd,fvals)
    plt.plot(np.log10(fvals),np.log10(Sh),label="Sh:iLIGO")
    Sh = map(lalsim.SimNoisePSDaLIGOZeroDetHighPower,fvals)
    plt.plot(np.log10(fvals),np.log10(Sh),label="Sh:aLIGO")
    plt.legend()
    plt.ylim(-50,-30)
    plt.xlim(0,3.5)
    plt.savefig("test-Q-response-psd.pdf")


TestDictionary = factored_likelihood_test.TestDictionaryDefault
TestDictionary["UVReport"]              =  analytic_signal  # this report is very confusing for real data
TestDictionary["QSquaredTimeseries"] = False # opts.plot_ShowLikelihoodVersusTime    # should be command-line option to control this plot specifically
TestDictionary["Rho22Timeseries"]      = True
TestDictionary["lnLModelAtKnown"]  =  analytic_signal  # this report is very confusing for real data
TestDictionary["lnLDataAtKnownPlusOptimalTimePhase"] = False
TestDictionary["lnLAtKnown"]           = True
TestDictionary["lnLAtKnownMarginalizeTime"]  = False
TestDictionary["AbsolutelyNoPlots"]   = not opts.plot_ShowLikelihoodVersusTime and not opts.plot_ShowSamplerInputs and not opts.plot_ShowSampler
TestDictionary["lnLDataPlot"]            = opts.plot_ShowLikelihoodVersusTime    # Plot individual geocentered L_k(t) and total L(t) [interpolated code]; plot discrete-data L_k(t)
TestDictionary["lnLDataPlotVersusPsi"] = opts.plot_ShowLikelihoodVersusTime
TestDictionary["lnLDataPlotVersusPhi"] = opts.plot_ShowLikelihoodVersusTime
TestDictionary["lnLDataPlotVersusPhiPsi"] = opts.plot_ShowLikelihoodVersusTime
print(" Detectors ", detectors)
factored_likelihood_test.TestLogLikelihoodInfrastructure(TestDictionary,theEpochFiducial,  data_dict, psd_dict,fmaxSNR, analyticPSD_Q, Psig, rholms,rholms_intp, crossTerms, detectors,Lmax)
